unit GsMemo;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls,
  ExtCtrls, GsRegExp;

type
  TGsCustomMemo = class;
  TGsAbout = type string;

  TGsMemoEnvironment = class(TPersistent)
  private
    FOwner: TGsCustomMemo;
    FclText: TColor;
    FclTextBk: TColor;
    FclSelection: TColor;
    FclSelectionBk: TColor;
    FclRightEdge: TColor;
    FclHyperlink: TColor;
    FclHyperlinkBk: TColor;
  protected
    procedure Set_clText(Value: TColor);
    procedure Set_clTextBk(Value: TColor);
    procedure Set_clSelection(Value: TColor);
    procedure Set_clSelectionBk(Value: TColor);
    procedure Set_clRightEdge(Value: TColor);
    procedure Set_clHyperlink(Value: TColor);
    procedure Set_clHyperlinkBk(Value: TColor);
  public
    constructor CreateEnvironment(AOwner: TGsCustomMemo);
    procedure Assign(Source: TPersistent); override;
    procedure SetAsDefault;
  published
    property Hyperlink: TColor read FclHyperlink write Set_clHyperlink;
    property HyperlinkBackground: TColor read FclHyperlinkBk write Set_clHyperlinkBk;
    property SelectionBackground: TColor read FclSelectionBk write Set_clSelectionBk;
    property SelectionColor: TColor read FclSelection write Set_clSelection;
    property TextBackground: TColor read FclTextBk write Set_clTextBk;
    property TextColor: TColor read FclText write Set_clText;
    property RightEdgeColor: TColor read FclRightEdge write Set_clRightEdge;
  end;

  TGsGutter = class(TCustomControl)
  private
    FOwner: TGsCustomMemo;
    FShowNum: Boolean;
    FStartNum: Integer;
    FBevel, FSelLine: Boolean;
    FAlignment: TAlignment;
    FSpace: Integer;
    FZeroStart: Boolean;
    FDigits: Integer;
    procedure SetShowNum(Value: Boolean);
    procedure SetStartNum(Value: Integer);
    procedure SetBevel(Value: Boolean);
    procedure SetAlignment(Value: TAlignment);
    procedure SetSpace(Value: Integer);
    procedure SetZeroStart(Value: Boolean);
    procedure SetDigits(Value: Integer);
  protected
    procedure Paint; override;
    procedure DrawLineNums;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  public
    constructor CreateGutter(AOwner: TGsCustomMemo);
    property Bevel: Boolean read FBevel write SetBevel default True;
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property Color;
    property LeadingDigits: Integer read FDigits write SetDigits default 0;
    property ShowNum: Boolean read FShowNum write SetShowNum default False;
    property Space: Integer read FSpace write SetSpace default 2;
    property StartNum: Integer read FStartNum write SetStartNum default 1;
    property Width;
    property ZeroStart: Boolean read FZeroStart write SetZeroStart;
  end;

  TGsGutterOptions = class(TPersistent)
  private
    FOwner: TGsCustomMemo;
    FVisible: Boolean;
    function GetGutterLineNum: Boolean;
    procedure SetGutterLineNum(Value: Boolean);
    function GetGutterBevel: Boolean;
    procedure SetGutterBevel(Value: Boolean);
    function GetGutterColor: TColor;
    procedure SetGutterColor(Value: TColor);
    function GetGutterFontColor: TColor;
    procedure SetGutterFontColor(Value: TColor);
    function GetGutterAlign: TAlignment;
    procedure SetGutterAlign(Value: TAlignment);
    procedure SetVisible(Value: Boolean);
    function GetGutterWidth: Integer;
    procedure SetGutterWidth(Value: Integer);
    function GetGutterSpace: Integer;
    procedure SetGutterSpace(Value: Integer);
    function GetGutterDigits: Integer;
    procedure SetGutterDigits(Value: Integer);
    function GetGutterZeroStart: Boolean;
    procedure SetGutterZeroStart(Value: Boolean);
  public
    constructor CreateGutterOptions(AOwner: TGsCustomMemo);
  published
    property Alignment: TAlignment read GetGutterAlign write SetGutterAlign;
    property Bevel: Boolean read GetGutterBevel write SetGutterBevel;
    property Color: TColor read GetGutterColor write SetGutterColor;
    property FontColor: TColor read GetGutterFontColor write SetGutterFontColor;
    property ShowLineNum: Boolean read GetGutterLineNum write SetGutterLineNum;
    property Space: Integer read GetGutterSpace write SetGutterSpace;
    property Visible: Boolean read FVisible write SetVisible;
    property Width: Integer read GetGutterWidth write SetGutterWidth;
    property LedingDigits: Integer read GetGutterDigits write SetGutterDigits;
    property ZeroStart: Boolean read GetGutterZeroStart write SetGutterZeroStart;
  end;

  TGsKeyEvent = (k_Down, k_Press, k_Up, k_Mouse1, k_Mouse2, k_Mouse3, k_DoubleClick, k_None);
  TGsScrolls = (sbNone, sbBoth, sbHorizontal, sbVertical);

  TGsFindOptions = set of (foMatchCase, foWholeWords, foEntireScope);

  TTextScroledEvent = procedure(Sender: TObject; TopRow, BottomRow,
    LeftCol, MaxCols: Integer) of object;
  TGsDrawLineEvent = procedure(Sender: TObject; LineText: string; LineIndxe: Integer;
    LineGlyph: TBitmap; TextRect: TRect; var AllowSelfDraw: Boolean) of object;
  TGsHyperlinkClickEvent = procedure(Sender: TObject; Hyperlink: string) of object;
  TGsMemoEvents = procedure(Sender: TObject; Shift: TShiftState; KeyEvent: TGsKeyEvent;
    Key: Word; CaretPos, MousePos: TPoint) of object;
  TGsFindText = procedure(Sender: TObject; FindPos: TPoint;
    var ScrollToWord: Boolean) of Object;
  TGsReplaceText = procedure(Sender: TObject; FindPos: TPoint;
    var ScrollToWord, ReplaceText: Boolean) of Object;
  TGsOnSearchEnd = procedure(Sender: TObject; FindIt, ReplaceMode: Boolean) of object;

  TGsCustomMemo = class(TCustomControl)
  private
    FWindowHandle: HWND;
    FTracking: Boolean;
    FFullRedraw: Boolean;
    FVScrollVisible, FHScrollVisible: Boolean;
    FOnScrolled_V, FOnScrolled_H: TNotifyEvent;
    FOnTextScrolled: TTextScroledEvent;
    FOnMouseWheel: TMouseWheelEvent;
    FOnMouseWheelDown: TMouseWheelUpDownEvent;
    FOnMouseWheelUp: TMouseWheelUpDownEvent;
    FOnPaint: TNotifyEvent;
    FScrollBars: TGsScrolls;
    FBorderStyle: TBorderStyle;
    FMaxScrollH: Integer;
    FWheelAccumulator: Integer;
    FRightEdge: Boolean;
    FRightEdgeCol: Integer;
    FShowHyperlinks: Boolean;
    FGutterPan: TGsGutter;
    FEnvironment: TGsMemoEnvironment;
    FGutterOpt: TGsGutterOptions;
    FLines: TStrings;
    FTabWidth: Integer;
    FChrW, FChrH: Integer;
    FTopLine, FVisLines, FVisCols: Integer;
    FSelStartNo, FSelEndNo, FSelStartOffs, FSelEndOffs: Integer;
    FSelected, FSelMouseDwn: Boolean;
    VScrollDelta, HScrollDelta, XMouse, YMouse: Integer;
    ScrollTimer: TTimer;
    HPos, VPos, XSize, YSize: Integer;
    StartNo, EndNo, StartOffs, EndOffs: Integer;
    CaretPos, MousePos: TPoint;
    ShiftState: TShiftState;
    HideCaret, PulseCaret, FDblClick, FSelDrag, FNotHideHint: Boolean;
    FModified: Boolean;
    REx: TRegExpCxMemo;
    FBuffer: TBitmap;
    FHintWnd: THintWindow;
    FDropList: TListBox;
    FDropList_CharW, FDropList_CharH: Integer;
    FDropTime, FDropTimeCount, FDropCol_CharW: Integer;
    FLastDropPos: TPoint;
    FDropListEvent: TNotifyEvent;
        
    FOnDrawLine: TGsDrawLineEvent;
    FOnHyperlinkClick: TGsHyperlinkClickEvent;
    FOnMemoEvents: TGsMemoEvents;
    FOnFindText: TGsFindText;
    FOnReplaceText: TGsReplaceText;
    FOnSerachEnd: TGsOnSearchEnd;

    FAbout: TGsAbout;

    function GetVScrollPos: Integer;
    procedure SetVScrollPos(Pos: Integer);
    function GetVScrollMax: Integer;
    function GetHScrollPos: Integer;
    procedure SetHScrollPos(Pos: Integer);
    function GetHScrollMax: Integer;

    procedure SetScrollBars(Value: TGsScrolls);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetLines(Value: TStrings);
    procedure SetShowHyperlinks(Value: Boolean);
    procedure SetTabWidth(Value: Integer);
    procedure SetRightEdge(Value: Boolean);
    procedure SetRightEdgeCol(Value: Integer);
    procedure CaretBlinkingTimeProc(var Msg: TMessage);
    function GetCaretVPos: Integer;
    function GetCaretHPos: Integer;
    procedure SetDropTime(Value: Integer);
    function GetDropList: TStrings;
    procedure SetDropList(Value: TStrings);
    function GetSelBegin: TPoint;
    function GetSelEnd: TPoint;
    procedure SetSelBegin(Value: TPoint);
    procedure SetSelEnd(Value: TPoint);
  protected
    procedure InitMemoObjects;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure UpdateTimer;
    procedure UpdateScrollBars;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMMouseWheel(var Message: TCMMouseWheel); message CM_MOUSEWHEEL;
    procedure MouseWheel(Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint); dynamic;
    function MouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; dynamic;
    function MouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; dynamic;
    function GetHyperlinksAtPos(X, Y: Integer; var SLink: string): Boolean;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DragDropSelection(AtPos: TPoint; CopyMode: Boolean);
    procedure DblClick; override;
    procedure SetVPos(p: Integer);
    procedure SetHPos(p: Integer);
    procedure ScrollChildren(dx, dy: Integer);
    procedure UpdateChildren;
    procedure UpdateCharBounds;
    procedure MarkHyperlinks(S: string; var LGliph: TBitmap);
    procedure FormatLine(I: Integer; var LGliph: TBitmap);
    procedure Paint; override;
    procedure DrawLine(CanvasSupport: TCanvas; I, Y: Integer);
    procedure DrawVisible;
    procedure GetSelBounds(var StartNo, EndNo, StartOffs, EndOffs: Integer);
    procedure ScrollOnTimer(Sender: TObject);
    procedure BlinkCaret;
    procedure DrawCaret(X, Y: Integer; ShowCaret: Boolean);
    procedure KeyboardCaretNav(ShiftState: TShiftState; Direction: Word);
    procedure ShowTextHint(X, Y: Integer; HintMsg: string);
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure WMKeyUp(var Message: TWMKeyUp); message WM_KEYUP;
    procedure KeyPress(var Key: Char); override;
    procedure DropListKeyPress(Sender: TObject; var Key: Char);
    procedure DropListExit(Sender: TObject);
    procedure DropListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
      State: TOwnerDrawState);
    function FindReplaceProc(TextToFind: String; FindOptions: TGsFindOptions;
      Backward, ReplaceMode: Boolean; var ReplaceText: Boolean): Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetSelText: string;
    function GetWordAtPos(Col, Row: Integer; var XBegin: Integer): string;
    procedure GetRowColAtPos(X, Y: Integer; var Row, Col: Integer);
    procedure InsertTextAtPos(S: string; Col, Row: Integer);
    procedure ScrollTo(X, Y: Integer);
    procedure ShowDropWindow;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure PasteFromClipboard;
    procedure DeleteSelection;
    procedure FindText(TextToFind: String; FindOptions: TGsFindOptions; Backward: Boolean);
    procedure ReplaceText(TextToFind, TextToReplace: String; FindOptions: TGsFindOptions; Backward: Boolean);
    procedure Clear;
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);
    property SelBegin: TPoint read GetSelBegin write SetSelBegin;
    property SelEnd: TPoint read GetSelEnd write SetSelEnd;
    property Selected: Boolean read FSelected;    
    property CaretPos_H: Integer read GetCaretHPos;
    property CaretPos_V: Integer read GetCaretVPos;
    property FontHeight: Integer read FChrH;
    property FontWidth: Integer read FChrW;
    property HideHintWindow: Boolean read FNotHideHint write FNotHideHint;
    property HintWindow: THintWindow read FHintWnd write FHintWnd;
    property Modified: Boolean read FModified write FModified;
    property ScrollBars: TGsScrolls read FScrollBars write SetScrollBars default sbBoth;
    property ScrollMax_H: Integer read GetHScrollMax;
    property ScrollMax_V: Integer read GetVScrollMax;
    property ScrollPos_H: Integer read GetHScrollPos write SetHScrollPos;
    property ScrollPos_V: Integer read GetVScrollPos write SetVScrollPos;
    property TopLine: Integer read FTopLine;
    property VisibleLines: Integer read FVisLines;
    { Published declarations }
    property About: TGsAbout read FAbout write FAbout;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property DropListCharHeight: Integer read FDropList_CharH write FDropList_CharH;
    property DropListCharWidth: Integer read FDropList_CharW write FDropList_CharW;
    property DropListItems: TStrings read GetDropList write SetDropList;
    property DropListTimeDelay: Integer read FDropTime write SetDropTime default 0;
    property DropListColumnCharWidth: Integer read FDropCol_CharW write FDropCol_CharW;
    property EnvironmentOptions: TGsMemoEnvironment read FEnvironment write FEnvironment;
    property FullRedraw: Boolean read FFullRedraw write FFullRedraw default False;
    property GutterOptions: TGsGutterOptions read FGutterOpt write FGutterOpt;
    property Lines: TStrings read FLines write SetLines;
    property RightEdge: Boolean read FRightEdge write SetRightEdge default False;
    property RightEdgeCol: Integer read FRightEdgeCol write SetRightEdgeCol default 80;
    property ShowHyperlinks: Boolean read FShowHyperlinks write SetShowHyperlinks default False;
    property TabWidth: Integer read FTabWidth write SetTabWidth;
    property Tracking: Boolean read FTracking write FTracking default False;
    property OnDrawLine: TGsDrawLineEvent read FOnDrawLine write FOnDrawLine;
    property OnDropListDown: TNotifyEvent read FDropListEvent write FDropListEvent;
    property OnFindText: TGsFindText read FOnFindText write FOnFindText;
    property OnHyperlinkClick: TGsHyperlinkClickEvent read FOnHyperlinkClick write FOnHyperlinkClick;
    property OnMemoEvents: TGsMemoEvents read FOnMemoEvents write FOnMemoEvents;
    property OnMouseWheel: TMouseWheelEvent read FOnMouseWheel write FOnMouseWheel;
    property OnMouseWheelDown: TMouseWheelUpDownEvent read FOnMouseWheelDown write FOnMouseWheelDown;
    property OnMouseWheelUp: TMouseWheelUpDownEvent read FOnMouseWheelUp write FOnMouseWheelUp;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnReplaceText: TGsReplaceText read FOnReplaceText write FOnReplaceText;
    property OnScrolled_H: TNotifyEvent read FOnScrolled_H write FOnScrolled_H;
    property OnScrolled_V: TNotifyEvent read FOnScrolled_V write FOnScrolled_V;
    property OnSearchEnd: TGsOnSearchEnd read FOnSerachEnd write FOnSerachEnd;
    property OnTextScrolled: TTextScroledEvent read FOnTextScrolled write FOnTextScrolled;
  end;

  TGsMemo = class(TGsCustomMemo)
  published
    property About;
    property Align;
    property BorderStyle;
    property Color;
    property Cursor;
    property DragCursor;
    property DragMode;
    property DropListCharHeight;
    property DropListCharWidth;
    property DropListItems;
    property DropListTimeDelay;
    property DropListColumnCharWidth;
    property Enabled;
    property EnvironmentOptions;
    property Font;
    property FullRedraw;
    property GutterOptions;
    property HelpContext;
    property Hint;
    property Lines;
    property ParentShowHint;
    property ParentColor;
    property RightEdge;
    property RightEdgeCol;
    property ShowHint;
    property ShowHyperlinks;
    property TabStop;
    property TabOrder;
    property TabWidth;
    property Tracking;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawLine;
    property OnDropListDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnFindText;
    property OnHyperlinkClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMemoEvents;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnPaint;
    property OnReplaceText;
    property OnResize;
    property OnScrolled_H;
    property OnScrolled_V;
    property OnSearchEnd;
    property OnStartDrag;
    property OnTextScrolled;
  end;


{$R GSMemo.res}

var
  RegularURL: string = '([Ff][Tt][Pp]|[Hh][Tt][Tt][Pp])://([_a-zA-Z\d\-]+(\.[_a-zA-Z\d\-]+)+)((/[_a-zA-Z\d\-\\\.\?\&\%\\~=]+)+)*';
  RegularEmail: string = '[_a-zA-Z\d\-\.]+@[_a-zA-Z\d\-]+(\.[_a-zA-Z\d\-]+)+';

const
  clSelColor = clHighlightText;
  clSelColorBk = clHighlight;
  clRightEdge = clGray;
  clGutter = clBtnFace;
  clGutterFont = clBtnText;
  clTextColor = clWindowText;
  clTextColorBk = clWindow;
  clHyperlink = clBlue;
  clHyperlinkBk = clWindow;
  crDragSel = 1;
  crDragSelCopy = 2;
  crHand = 3;
  crSelLine = 4;
  DT_DRAWLINE = DT_NOPREFIX or DT_LEFT or DT_SINGLELINE or DT_VCENTER;

{+#NF}

  CARET_TIME_ERROR   = 'Caret timer error';
  S_IMAGE_HINT       = 'Image: ';
  S_LINK_HINT        = 'Link to: ';

{-#NF}

implementation

uses
  Clipbrd;

var
  SelDragCursor1, SelDragCursor2, HandCursor, SelLineCursor: TCursor;
  SavedType: String;
  RepPos: TPoint;

(***)

procedure CtrlByTagUpdate(AControl: TControl);
begin
  if AControl.Tag > 10000 then
    AControl.Top := 10000
  else
    if AControl.Tag < -10000 then
      AControl.Top := -10000
    else
      AControl.Top := AControl.Tag;
end;

(* TGsMemoEnvironment ***********************************************************************)

constructor TGsMemoEnvironment.CreateEnvironment(AOwner: TGsCustomMemo);
begin
  inherited Create;
  FOwner := AOwner;
  FclText := clTextColor;
  FclTextBk := clTextColorBk;
  FclSelection := clSelColor;
  FclSelectionBk := clSelColorBk;
  FclRightEdge := clRightEdge;
  FclHyperlink := clHyperlink;
  FclHyperlinkBk := clHyperlinkBk;
end;

procedure TGsMemoEnvironment.Set_clText(Value: TColor);
begin
  if FclText <> Value then
  begin
    FclText := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Set_clTextBk(Value: TColor);
begin
  if FclTextBk <> Value then
  begin
    FclTextBk := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Set_clSelection(Value: TColor);
begin
  if FclSelection <> Value then
  begin
    FclSelection := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Set_clSelectionBk(Value: TColor);
begin
  if FclSelectionBk <> Value then
  begin
    FclSelectionBk := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Set_clRightEdge(Value: TColor);
begin
  if FclRightEdge <> Value then
  begin
    FclRightEdge := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Set_clHyperlink(Value: TColor);
begin
  if FclHyperlink <> Value then
  begin
    FclHyperlink := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Set_clHyperlinkBk(Value: TColor);
begin
  if FclHyperlinkBk <> Value then
  begin
    FclHyperlinkBk := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

procedure TGsMemoEnvironment.Assign(Source: TPersistent);
var
  Src: TGsMemoEnvironment;
begin
  if Source is TGsMemoEnvironment then
  begin
    Src := TGsMemoEnvironment(Source);
    FclText := Src.TextColor;
    FclTextBk := Src.TextBackground;
    FclSelection := Src.SelectionColor;
    FclSelectionBk := Src.SelectionBackground;
    FclRightEdge := Src.RightEdgeColor;
    FclHyperlink := Src.Hyperlink;
    FclHyperlinkBk := Src.HyperlinkBackground;
  end else
    inherited Assign(Source);
end;

procedure TGsMemoEnvironment.SetAsDefault;
begin
  FclText := clTextColor;
  FclTextBk := clTextColorBk;
  FclSelection := clSelColor;
  FclSelectionBk := clSelColorBk;
  FclRightEdge := clRightEdge;
  if FOwner <> nil then FOwner.Invalidate;
end;

(* TGsGutter ********************************************************************************)

constructor TGsGutter.CreateGutter(AOwner: TGsCustomMemo);
begin
  inherited Create(AOwner);
  FOwner := AOwner;
  FShowNum := False;
  FStartNum := 1;
  FAlignment := taLeftJustify;
  FSpace := 2;
  FBevel := True;
  ParentColor := False;
  ParentShowHint := False;
  Color := clGutter;
  Font.Color := clGutterFont;
  Width := 26;
  SelLineCursor := LoadCursor(hInstance, 'LINESEL');
  Screen.Cursors[crSelLine] := SelLineCursor;
end;

procedure TGsGutter.SetShowNum(Value: Boolean);
begin
  if FShowNum <> Value then
  begin
    FShowNum := Value;
    Invalidate;
  end;
end;

procedure TGsGutter.SetStartNum(Value: Integer);
begin
  if Value < 1 then Value := 1;
  FStartNum := Value;
end;

procedure TGsGutter.SetBevel(Value: Boolean);
begin
  if FBevel <> Value then
  begin
    FBevel := Value;
    Invalidate;
  end;
end;

procedure TGsGutter.SetSpace(Value: Integer);
begin
  if Value < 1 then Value := 1;
  if Value > (Width div 3) then Value := Width div 3;
  if FSpace <> Value then
  begin
    FSpace := Value;
    Invalidate;
  end;
end;

procedure TGsGutter.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    Invalidate;
  end;
end;

procedure TGsGutter.SetZeroStart(Value: Boolean);
begin
  if FZeroStart <> Value then
  begin
    FZeroStart := Value;
    Invalidate;
  end;
end;

procedure TGsGutter.SetDigits(Value: Integer);
begin
  if Value < 0 then Value := 0;
  if Value > 8 then Value := 8;
  if FDigits <> Value then
  begin
    FDigits := Value;
    Invalidate;
  end;
end;

procedure TGsGutter.Paint;
var
  W: Integer;
begin
  if FSpace > (Width div 3) then FSpace := Width div 3;
  if Width < 14 then Width := 14;
  with Canvas do
  begin
    Brush.Color := Self.Color;
    FillRect(ClientRect);
    W := ClientWidth;
    if FOwner <> nil then
    begin
      Pen.Color := FOwner.Color;
      MoveTo(W - 1, 0);
      LineTo(W - 1, ClientHeight);
    end;
    if FBevel then
    begin
      Pen.Color := clBtnHighlight;
      MoveTo(W - 3, 0);
      LineTo(W - 3, ClientHeight);
      Pen.Color := clBtnShadow;
      MoveTo(W - 2, 0);
      LineTo(W - 2, ClientHeight);
    end;
  end;
  DrawLineNums;
end;

procedure TGsGutter.DrawLineNums;
var
  RC: TRect;
  I, MaxI, W, H, ZeroL: Integer;
  DT_Params: Word;
  S: string;
begin
  if not FShowNum then Exit;
  W := ClientWidth - FSpace - 1;
  if FBevel then W := W - 3;
  with Canvas do
  begin
    if FOwner <> nil then
    begin
      Font.Name := 'MS Sans Serif';
      Font.Size := 8;
      Font.Color := Self.Font.Color;
      Brush.Color := Self.Color;
      H := FOwner.FChrH;
      MaxI := FOwner.FVisLines;
      DT_Params := DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX;
      case FAlignment of
        taLeftJustify: DT_Params := DT_LEFT or DT_Params;
        taCenter: DT_Params := DT_CENTER or DT_Params;
        taRightJustify: DT_Params := DT_RIGHT or DT_Params;
      end;
      for I := 0 to MaxI do
      begin
        if FZeroStart then
          S := IntToStr(FStartNum + I - 1)
        else
          S := IntToStr(FStartNum + I);
        for ZeroL := Length(S) to FDigits do S := '0' + S;
        RC := Rect(2, I * H, W, (I * H) + H);
        FillRect(RC);
        DrawText(Canvas.Handle, PChar(S), Length(S), RC, DT_Params);
      end;
    end;
  end;
end;

procedure TGsGutter.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  SN, EN, SO, EO, RNo, CNo: Integer;
  S: string;
begin
  inherited MouseUp(Button, Shift, X, Y);
  FSelLine := False;
  if (X < Width - 1) and (X > 2 * (Width div 3)) then FSelLine := True;
  if FSelLine and (FOwner <> nil) then
    with FOwner do
    begin
      GetRowColAtPos(X, Y, RNo, CNo);
      S := '';
      if RNo <= FLines.Count - 1 then S := FLines[RNo];
      if ssShift in Shift then
      begin
        GetSelBounds(SN, EN, SO, EO);
        if RNo < SN then
        begin
          FSelStartNo := RNo;
          FSelStartOffs := 0;
          FSelEndNo := EN;
          FSelEndOffs := EO;
          CaretPos.x := RNo;
          CaretPos.y := 0;
          DrawCaret(CaretPos.x, CaretPos.y, False);
          FSelected := True;
          DrawVisible;
          SetFocus;
        end else
          if RNo > EN then
          begin
            FSelStartNo := SN;
            FSelStartOffs := 0;
            FSelEndNo := RNo;
            FSelEndOffs := Length(S);
            CaretPos.x := RNo;
            CaretPos.y := 0;
            DrawCaret(CaretPos.x, CaretPos.y, False);
            FSelected := True;
            DrawVisible;
            SetFocus;
          end;
      end else
      begin
        FSelStartNo := RNo;
        FSelEndNo := RNo;
        CaretPos.x := RNo;
        CaretPos.y := 0;
        FSelStartOffs := 0;
        FSelEndOffs := Length(S);
        DrawCaret(CaretPos.x, CaretPos.y, False);
        FSelected := (FSelStartOffs <> FSelEndOffs);
        DrawVisible;
        SetFocus;
      end;
    end;
end;

procedure TGsGutter.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  FSelLine := False;
  if (X < Width - 1) and (X > 2 * (Width div 3)) then FSelLine := True;
  if FSelLine then
    Cursor := crSelLine
  else
    Cursor := crDefault;
  Screen.Cursor := Cursor;
end;

(* TGsGutterOptions ***************************************************************************)

constructor TGsGutterOptions.CreateGutterOptions(AOwner: TGsCustomMemo);
begin
  inherited Create;
  FOwner := AOwner;
  FVisible := True;
end;

function TGsGutterOptions.GetGutterSpace: Integer;
begin
  Result := 1;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.Space;
end;

procedure TGsGutterOptions.SetGutterSpace(Value: Integer);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      FOwner.FGutterPan.Space := Value;
end;

function TGsGutterOptions.GetGutterDigits: Integer;
begin
  Result := 1;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.LeadingDigits;
end;

procedure TGsGutterOptions.SetGutterDigits(Value: Integer);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      FOwner.FGutterPan.LeadingDigits := Value;
end;

function TGsGutterOptions.GetGutterZeroStart: Boolean;
begin
  Result := False;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.ZeroStart;
end;

procedure TGsGutterOptions.SetGutterZeroStart(Value: Boolean);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      FOwner.FGutterPan.ZeroStart := Value;
end;

procedure TGsGutterOptions.SetVisible(Value: Boolean);
begin
  FVisible := Value;
  if FOwner <> nil then FOwner.Invalidate;
end;

function TGsGutterOptions.GetGutterAlign: TAlignment;
begin
  Result := taLeftJustify;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.Alignment;
end;

procedure TGsGutterOptions.SetGutterAlign(Value: TAlignment);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      FOwner.FGutterPan.Alignment := Value;
end;

function TGsGutterOptions.GetGutterWidth: Integer;
begin
  Result := 16;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.Width;
end;

procedure TGsGutterOptions.SetGutterWidth(Value: Integer);
begin
  if Value < 12 then Value := 12;
  if Value > 48 then Value := 48;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
    begin
      FOwner.FGutterPan.Width := Value;
      FOwner.Invalidate;
    end;
end;

function TGsGutterOptions.GetGutterColor: TColor;
begin
  Result := clNone;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.Color;
end;

procedure TGsGutterOptions.SetGutterColor(Value: TColor);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      if FOwner.FGutterPan.Color <> Value then
        FOwner.FGutterPan.Color := Value;
end;

function TGsGutterOptions.GetGutterFontColor: TColor;
begin
  Result := clNone;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.Font.Color
    else
      Result := clNone;
end;

procedure TGsGutterOptions.SetGutterFontColor(Value: TColor);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      if FOwner.FGutterPan.Font.Color <> Value then
      begin
        FOwner.FGutterPan.Font.Color := Value;
        FOwner.FGutterPan.Invalidate;
      end;
end;

function TGsGutterOptions.GetGutterLineNum: Boolean;
begin
  Result := False;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.ShowNum
end;

procedure TGsGutterOptions.SetGutterLineNum(Value: Boolean);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan.ShowNum <> Value then
      FOwner.FGutterPan.ShowNum := Value;
end;

function TGsGutterOptions.GetGutterBevel: Boolean;
begin
  Result := False;
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      Result := FOwner.FGutterPan.Bevel;
end;

procedure TGsGutterOptions.SetGutterBevel(Value: Boolean);
begin
  if FOwner <> nil then
    if FOwner.FGutterPan <> nil then
      FOwner.FGutterPan.Bevel := Value;
end;

(* TGsCustomMemo ******************************************************************************)

constructor TGsCustomMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable, csDisplayDragImage, csActionClient];
  FWindowHandle := AllocateHWnd(CaretBlinkingTimeProc);
  Width := 320;
  Height := 240;
  TabStop := True;
  FTracking := True;
  FFullRedraw := False;
  FScrollBars := sbNone;
  FVScrollVisible := True;
  FHScrollVisible := True;
  FBorderStyle := bsSingle;
  FLines := TStringList.Create;
  ParentColor := False;
  Color := clWindow;
  Font.Name := 'Fixedsys';
  Font.Color := clWindowText;
  Font.Size := 10;
  FTopLine := 0;
  FTabWidth := 8;
  FMaxScrollH := 0;
  Cursor := crIBeam;
  FShowHyperlinks := False;
  FRightEdge := False;
  FRightEdgeCol := 80;
  FSelected := False;
  FSelMouseDwn := False;
  CaretPos.x := 0;
  CaretPos.y := 0;
  PulseCaret := True;
  FDblClick := False;
  FSelDrag := False;
  FBuffer := TBitmap.Create;
  FDropTime := 3;
  FDropList_CharH := 4;
  FDropList_CharW := 20;
  FDropCol_CharW := 10;
  FLastDropPos.x := 0;
  FLastDropPos.y := 0;
  FAbout := '(About ' + ClassName + ')';
  InitMemoObjects;
end;

procedure TGsCustomMemo.InitMemoObjects;
begin
  FGutterPan := TGsGutter.CreateGutter(Self);
  with FGutterPan do
  begin
    Left := -Width - 1;
    Parent := Self;
    Visible := False;
  end;
  REx := TRegExpCxMemo.Create;
  FEnvironment := TGsMemoEnvironment.CreateEnvironment(Self);
  FGutterOpt := TGsGutterOptions.CreateGutterOptions(Self);
  SelDragCursor1 := LoadCursor(hInstance, 'SELDRAG');
  Screen.Cursors[crDragSel] := SelDragCursor1;
  SelDragCursor2 := LoadCursor(hInstance, 'SELDRAGCOPY');
  Screen.Cursors[crDragSelCopy] := SelDragCursor2;
  HandCursor := LoadCursor(hInstance, 'HANDPOINT');
  Screen.Cursors[crHand] := HandCursor;
  { Special Hint Window }
  FHintWnd := nil;
  { DropList Window }
  FDropList := TListBox.Create(Self);
  FDropList.Left := -200;
  FDropList.Parent := Self;
  FDropList.Visible := False;
  FDropList.ParentCtl3D := False;
  FDropList.Ctl3D := False;
  FDropList.IntegralHeight := True;
  FDropList.Sorted := True;
  FDropList.Style := lbOwnerDrawFixed;
  FDropList.OnKeyPress := DropListKeyPress;
  FDropList.OnExit := DropListExit;
  FDropList.OnDrawItem := DropListDrawItem;
end;

destructor TGsCustomMemo.Destroy;
begin
  FLines.Free;
  FEnvironment.Free;
  FGutterOpt.Free;
  FGutterPan.Free;
  if ScrollTimer <> nil then ScrollTimer.Free;
  KillTimer(FWindowHandle, 1);
  DeallocateHWnd(FWindowHandle);
  REx.Free;
  FBuffer.Free;
  if FHintWnd <> nil then FHintWnd.Free;
  FDropList.Free;
  inherited Destroy;
end;

procedure TGsCustomMemo.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    Style := Style or WS_CLIPCHILDREN or WS_HSCROLL or WS_VSCROLL;
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    WindowClass.Style := WindowClass.Style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TGsCustomMemo.CreateWnd;
begin
  inherited CreateWnd;
  VPos := 0;
  HPos := 0;
  UpdateCharBounds;
  UpdateScrollBars;
  ScrollTimer := TTimer.Create(Self);
  ScrollTimer.Interval := 100;
  ScrollTimer.Enabled := False;
  ScrollTimer.OnTimer := ScrollOnTimer;
  if csDesigning in ComponentState then
    if FLines.Count = 0 then FLines.Add(Name);
  UpdateTimer;
end;

procedure TGsCustomMemo.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result or DLGC_WANTTAB or DLGC_WANTARROWS;
end;

procedure TGsCustomMemo.CaretBlinkingTimeProc(var Msg: TMessage);
begin
  with Msg do
    if Msg = WM_TIMER then
    try
      BlinkCaret;
    except
      Application.HandleException(Self);
    end
    else
      Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

procedure TGsCustomMemo.UpdateTimer;
begin
  KillTimer(FWindowHandle, 1);
  if PulseCaret then
    if SetTimer(FWindowHandle, 1, 450, nil) = 0 then
      raise EOutOfResources.Create(CARET_TIME_ERROR);
end;

{ Setting propertyes }

procedure TGsCustomMemo.SetRightEdge(Value: Boolean);
begin
  if FRightEdge <> Value then
  begin
    FRightEdge := Value;
    Invalidate;
  end;
end;

procedure TGsCustomMemo.SetRightEdgeCol(Value: Integer);
begin
  if Value < 20 then Value := 20;
  if Value > 160 then Value := 160;
  if FRightEdgeCol <> Value then
  begin
    FRightEdgeCol := Value;
    if FRightEdge then Invalidate;
  end;
end;

procedure TGsCustomMemo.SetScrollBars(Value: TGsScrolls);
const
  V_SCROLL: array[TGsScrolls] of Boolean = (False, True, False, True);
  H_SCROLL: array[TGsScrolls] of Boolean = (False, True, True, False);
begin
  if Value <> FScrollBars then
  begin
    FScrollBars := Value;
    FVScrollVisible := V_SCROLL[Value];
    FHScrollVisible := H_SCROLL[Value];
    ShowScrollBar(Handle, SB_VERT, FVScrollVisible);
    ShowScrollBar(Handle, SB_HORZ, FHScrollVisible);
  end;
end;

procedure TGsCustomMemo.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TGsCustomMemo.SetLines(Value: TStrings);
begin
  FLines.Assign(Value);
  Invalidate;
end;

procedure TGsCustomMemo.SetShowHyperlinks(Value: Boolean);
begin
  if FShowHyperlinks <> Value then
  begin
    FShowHyperlinks := Value;
    Invalidate;
  end;
end;

procedure TGsCustomMemo.SetTabWidth(Value: Integer);
begin
  if Value < 1 then
  begin
    if csDesigning in ComponentState then
      MessageBox(Application.Handle, 'Value for TabWidth must be greater than 0.',
       PChar(ClassName + ' Tip'), MB_OK or MB_ICONINFORMATION);
    Value := FTabWidth;
  end;
  FTabWidth := Value;
end;

function TGsCustomMemo.GetCaretVPos: Integer;
begin
  Result := CaretPos.x;
end;

function TGsCustomMemo.GetCaretHPos: Integer;
begin
  Result := CaretPos.y;
end;

function TGsCustomMemo.GetDropList: TStrings;
begin
  Result := FDropList.Items;
end;

procedure TGsCustomMemo.SetDropList(Value: TStrings);
begin
  FDropList.Items.Assign(Value);
end;

{ DropDown List Implementation }

procedure TGsCustomMemo.SetDropTime(Value: Integer);
begin
  FDropTime := Value;
  if FDropTime < 0 then FDropTime := 0;
  if FDropTime > 10 then FDropTime := 10;
end;

procedure TGsCustomMemo.ShowDropWindow;
var
  Chk: Integer;
  S, SLine: String;
begin
  if Assigned(FDropListEvent) then FDropListEvent(Self);

  { Show DropList Window when: Component is not in Designing state, FDropList
    is not visible, Component is not in Selection state and FDropList is not
    empty (has items). }

  if csDesigning in ComponentState then Exit;
  if FDropList.Visible then Exit;
  if FSelected then Exit;
  if FDropList.Items.Count = 0 then Exit;
  FDropList.Visible := False;
  FDropList.Top := (CaretPos.x - FTopLine + 1) * FChrH;
  FDropList.Left := CaretPos.y * FChrW;
  if FDropList.ClientHeight <> ((FChrH * FDropList_CharH) + FChrH) then
    FDropList.ClientHeight := (FChrH * FDropList_CharH) + FChrH;
  if FDropList.ClientWidth <> ((FChrW * FDropList_CharW) + FChrW) then
    FDropList.ClientWidth := (FChrW * FDropList_CharW) + FChrW;
  if FGutterOpt.Visible then FDropList.Left := FDropList.Left + FGutterOpt.Width;
  Chk := FDropList.Top + FDropList.Height;
  if Chk > ClientHeight then FDropList.Top := FDropList.Top - FChrH - FDropList.Height;
  Chk := FDropList.Left + FDropList.Width;
  if Chk > ClientWidth then FDropList.Left := FDropList.Left - FDropList.Width;
  if CaretPos.x < FLines.Count then
  begin
    SLine := FLines[CaretPos.x];
    if (CaretPos.y) > Length(SLine) then Exit;
  end;
  S := GetWordAtPos(CaretPos.y, CaretPos.x, Chk);
  FDropList.ItemIndex := 0;
  FDropList.Visible := True;
  FLastDropPos.x := CaretPos.x;
  FLastDropPos.y := CaretPos.y;
  FDropList.SetFocus;
end;

procedure TGsCustomMemo.DropListKeyPress(Sender: TObject; var Key: Char);
var
  S, Cs: String;
  I: Integer;
  C1, C2: Char;
begin
  case Ord(Key) of
    VK_ESCAPE: begin
      FDropList.Visible := False;
      SavedType := '';
      SetFocus;
    end;
    VK_RETURN: begin
      FDropList.Visible := False;
      if FDropList.ItemIndex < 0 then
      begin
        SetFocus;
        SavedType := '';
        Exit;
      end;
      S := FDropList.Items[FDropList.ItemIndex];
      I := Pos('|', S);
      if I > 0 then
      begin
        Delete(S, 1, I);
        I := Pos('|', S);
        if I > 0 then Delete(S, 1, I);
        Cs := S;
      end else
        Cs := S;
      SavedType := Trim(SavedType);
      for I := 1 to Length(Cs) do
      begin
        if I <= Length(SavedType) then
        begin
          C1 := UpCase(SavedType[I]);
          C2 := UpCase(Cs[I]);
          if C1 = C2 then Continue;
        end;  
        KeyPress(Cs[I]);
      end;
      SavedType := '';
      SetFocus;
    end;
  else
    KeyPress(Key);
    SavedType := SavedType + Key;
  end;
end;

procedure TGsCustomMemo.DropListExit(Sender: TObject);
begin
  FDropList.Visible := False;
end;

procedure TGsCustomMemo.DropListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
  State: TOwnerDrawState);
var
  S, Cs1, Cs2: String;
  P: Integer;
begin
  FDropList.Canvas.Font.Name := 'MS Sans Serif';
  FDropList.Canvas.Font.Size := 8;
  FDropList.Canvas.Font.Style := [];
  if odSelected in State then
    FDropList.Canvas.Font.Color := clHighlightText
  else
    FDropList.Canvas.Font.Color := clWindowText;
  FDropList.Canvas.FillRect(Rect);
  S := FDropList.Items[Index];
  P := Pos('|', S);
  if P = 0 then
  begin
    S := ' ' + S;
    DrawText(FDropList.Canvas.Handle, PChar(S), Length(S), Rect, DT_DRAWLINE);
  end else
  begin
    Cs1 := '  ' + Copy(S, 1, P - 1);
    Delete(S, 1, P);
    P := Pos('|', S);
    if P > 0 then
      Cs2 := ' ' + Copy(S, 1, P - 1)
    else
      Cs2 := ' ' + S;
    DrawText(FDropList.Canvas.Handle, PChar(Cs2), Length(Cs2), Rect, DT_DRAWLINE);
    Rect.Left := (FChrW * FDropCol_CharW) + FChrW;
    FDropList.Canvas.Font.Style := [fsBold];    
    DrawText(FDropList.Canvas.Handle, PChar(Cs1), Length(Cs1), Rect, DT_DRAWLINE);
  end;
end;

{ Scrolling implementation }

procedure TGsCustomMemo.UpdateScrollBars;
var
  ScrollInfoV, ScrollInfoH: TScrollInfo;
begin
  ScrollInfoV.cbSize := SizeOf(ScrollInfoV);
  ScrollInfoV.fMask := SIF_ALL;
  ScrollInfoV.nMin := 0;
  ScrollInfoV.nPage := FVisLines - 2;
  ScrollInfoV.nMax := FLines.Count + 1;
  ScrollInfoV.nPos := VPos;
  ScrollInfoV.nTrackPos := 0;
  SetScrollInfo(Handle, SB_VERT, ScrollInfoV, True);

  ScrollInfoH.cbSize := SizeOf(ScrollInfoH);
  ScrollInfoH.fMask := SIF_ALL;
  ScrollInfoH.nMin := 0;
  ScrollInfoH.nPage := FMaxScrollH div 4;
  ScrollInfoH.nMax := FMaxScrollH;
  ScrollInfoH.nPos := HPos;
  ScrollInfoH.nTrackPos := 0;
  SetScrollInfo(Handle, SB_HORZ, ScrollInfoH, True);

  if not FVScrollVisible then ShowScrollBar(Handle, SB_VERT, FVScrollVisible);
  if not FHScrollVisible then ShowScrollBar(Handle, SB_HORZ, FHScrollVisible);
  UpdateChildren;
end;

procedure TGsCustomMemo.UpdateChildren;
var
  I: Integer;
begin
  for I := 0 to ControlCount - 1 do
  begin
    if (Controls[I] is TGsGutter) then
      if (TGsGutter(Controls[I]) = FGutterPan) then Continue;
    if (Controls[I] is TListBox) then
      if (TListBox(Controls[I]) = FDropList) then Continue;      
    CtrlByTagUpdate(Controls[I]);
  end;
end;

procedure TGsCustomMemo.ScrollChildren(dx, dy: Integer);
var
  I: Integer;
begin
  if (dx = 0) and (dy = 0) then Exit;
  for I := 0 to ControlCount - 1 do
  begin
    if (Controls[I] is TGsGutter) then
      if (TGsGutter(Controls[I]) = FGutterPan) then
        Continue;
    if dy <> 0 then
    begin
      Controls[I].Tag := Controls[I].Tag + dy;
      CtrlByTagUpdate(Controls[I]);
    end;
    if dx <> 0 then Controls[I].Left := Controls[I].Left + dx;
  end;
end;

procedure TGsCustomMemo.WMHScroll(var Message: TWMHScroll);
begin
  if FDropList.Visible then
    FDropList.Visible := False;
  FDropTimeCount := 0;
  FLastDropPos.x := -1;
  FLastDropPos.y := -1;
  with Message do
    case ScrollCode of
      SB_LINERIGHT: SetHPos(HPos + 1);
      SB_LINELEFT: SetHPos(HPos - 1);
      SB_PAGEUP: SetHPos(HPos - FVisLines);
      SB_PAGEDOWN: SetHPos(HPos + FVisLines);
      SB_THUMBPOSITION: SetHPos(Pos);
      SB_THUMBTRACK: if FTracking then SetHPos(Pos);
      SB_TOP: SetHPos(0);
      SB_BOTTOM: SetHPos(XSize);
    end;
end;

procedure TGsCustomMemo.WMVScroll(var Message: TWMVScroll);
begin
  if FDropList.Visible then
    FDropList.Visible := False;
  FDropTimeCount := 0;
  FLastDropPos.x := -1;
  FLastDropPos.y := -1;  
  with Message do
    case ScrollCode of
      SB_LINEUP: SetVPos(VPos - 1);
      SB_LINEDOWN: SetVPos(VPos + 1);
      SB_PAGEUP: SetVPos(VPos - FVisLines);
      SB_PAGEDOWN: SetVPos(VPos + FVisLines);
      SB_THUMBPOSITION: SetVPos(Pos);
      SB_THUMBTRACK: if FTracking then SetVPos(Pos);
      SB_TOP: SetVPos(0);
      SB_BOTTOM: SetVPos(YSize);
    end;
end;

procedure TGsCustomMemo.SetVPos(p: Integer);
var
  ScrollInfo: TScrollInfo;
  oldPos: Integer;
  Rc: TRect;
begin
  OldPos := VPos;
  VPos := p;
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.nPos := VPos;
  ScrollInfo.fMask := SIF_POS;
  SetScrollInfo(Handle, SB_VERT, ScrollInfo, True);
  GetScrollInfo(Handle, SB_VERT, ScrollInfo);
  VPos := ScrollInfo.nPos;
  Rc := ClientRect;
  if OldPos - VPos <> 0 then
  begin
    ScrollChildren(0, (OldPos - VPos) * FChrH);
    FTopLine := VPos;
    if FFullRedraw then
      Refresh
    else
      if (FTopLine + FVisLines) < Pred(FLines.Count) then
        DrawVisible
      else
        ScrollWindowEx(Handle, 0, (OldPos - VPos) * FChrH, nil, @Rc, 0, nil, SW_INVALIDATE);
    if Assigned(FOnScrolled_V) then FOnScrolled_V(Self);
    if Assigned(FOnTextScrolled) then FOnTextScrolled(Self, FTopLine, FTopLine + FVisLines + 1,
        HPos, FMaxScrollH);
  end;
end;

procedure TGsCustomMemo.SetHPos(p: Integer);
var
  ScrollInfo: TScrollInfo;
  oldPos: Integer;
  Rc: TRect;
begin
  OldPos := HPos;
  HPos := p;
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.nPos := HPos;
  ScrollInfo.fMask := SIF_POS;
  SetScrollInfo(Handle, SB_HORZ, ScrollInfo, True);
  GetScrollInfo(Handle, SB_HORZ, ScrollInfo);
  HPos := ScrollInfo.nPos;
  Rc := ClientRect;
  if OldPos - HPos <> 0 then
  begin
    ScrollChildren((OldPos - HPos), 0);
    if FFullRedraw then
      Refresh
    else
      DrawVisible;
    if Assigned(FOnScrolled_H) then FOnScrolled_H(Self);
    if Assigned(FOnTextScrolled) then FOnTextScrolled(Self, FTopLine, FTopLine + FVisLines,
        HPos, FMaxScrollH);
  end;
end;

procedure TGsCustomMemo.ScrollTo(X, Y: Integer);
begin
  SetVPos(Y div FChrH);
  SetHPos(X div FChrW);
end;

function TGsCustomMemo.GetVScrollPos: Integer;
begin
  Result := VPos;
end;

procedure TGsCustomMemo.SetVScrollPos(Pos: Integer);
begin
  SetVPos(Pos);
end;

function TGsCustomMemo.GetHScrollPos: Integer;
begin
  Result := HPos;
end;

procedure TGsCustomMemo.SetHScrollPos(Pos: Integer);
begin
  SetHPos(Pos);
end;

function TGsCustomMemo.GetVScrollMax: Integer;
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.nPos := HPos;
  ScrollInfo.fMask := SIF_RANGE or SIF_PAGE;
  GetScrollInfo(Handle, SB_VERT, ScrollInfo);
  Result := ScrollInfo.nMax - Integer(ScrollInfo.nPage - 1);
end;

function TGsCustomMemo.GetHScrollMax: Integer;
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.nPos := HPos;
  ScrollInfo.fMask := SIF_RANGE or SIF_PAGE;
  GetScrollInfo(Handle, SB_HORZ, ScrollInfo);
  Result := ScrollInfo.nMax - Integer(ScrollInfo.nPage - 1);
end;

{ Mouse Wheel scrolling implementation }

procedure TGsCustomMemo.CMMouseWheel(var Message: TCMMouseWheel);
begin
  with Message do
    MouseWheel(ShiftState, WheelDelta, SmallPointToPoint(Pos));
end;

procedure TGsCustomMemo.MouseWheel(Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint);
var
  ScrollNotify: Integer;
  hasShift, hasCtrl, NotDoIt, IsNeg: Boolean;
begin
  NotDoIt := False;
  if Assigned(FOnMouseWheel) then
    FOnMouseWheel(Self, Shift, WheelDelta, MousePos, NotDoIt);
  if NotDoIt then Exit;
  ScrollNotify := -1;
  hasShift := (ssShift in Shift);
  hasCtrl := (ssCtrl in Shift);
  if hasCtrl then
  begin
    if WheelDelta > 0 then ScrollNotify := SB_LINELEFT;
    if WheelDelta < 0 then ScrollNotify := SB_LINERIGHT;
    if (ScrollNotify <> -1) then
    begin
      Perform(WM_HSCROLL, ScrollNotify, 0);
      Perform(WM_HSCROLL, ScrollNotify, 0);
    end;
    Exit;
  end;
  if hasShift then
  begin
    DrawVisible;
    HideCaret := False;
    if WheelDelta > 0 then CaretPos.x := CaretPos.x - 1;
    if WheelDelta < 0 then CaretPos.x := CaretPos.x + 1;
    if CaretPos.x < 0 then CaretPos.x := 0;
    if CaretPos.x > pred(FLines.Count) then CaretPos.x := pred(FLines.Count);
    DrawCaret(CaretPos.x, CaretPos.y, HideCaret);
    Exit;
  end;
  if not hasShift and not hasCtrl then
  begin
    if WheelDelta > 0 then ScrollNotify := SB_LINEUP;
    if WheelDelta < 0 then ScrollNotify := SB_LINEDOWN;
    if (ScrollNotify <> -1) then Perform(WM_VSCROLL, ScrollNotify, 0);
  end;
  Inc(FWheelAccumulator, WheelDelta);
  while Abs(FWheelAccumulator) >= WHEEL_DELTA do
  begin
    IsNeg := FWheelAccumulator < 0;
    FWheelAccumulator := Abs(FWheelAccumulator) - WHEEL_DELTA;
    if IsNeg then
    begin
      if FWheelAccumulator <> 0 then FWheelAccumulator := -FWheelAccumulator;
      NotDoIt := MouseWheelDown(Shift, MousePos);
    end
    else
      NotDoIt := MouseWheelUp(Shift, MousePos);
  end;
end;

function TGsCustomMemo.MouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  Result := False;
  if Assigned(FOnMouseWheelDown) then FOnMouseWheelDown(Self, Shift, MousePos, Result);
end;

function TGsCustomMemo.MouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  Result := False;
  if Assigned(FOnMouseWheelUp) then FOnMouseWheelUp(Self, Shift, MousePos, Result);
end;

{ Control painting impementation }

procedure TGsCustomMemo.UpdateCharBounds;
begin
  Canvas.Font.Assign(Self.Font);
  FChrW := Canvas.TextWidth('W');
  FChrH := Canvas.TextHeight('Wp');
  FVisLines := (ClientHeight div FChrH) + 1;
  if FGutterOpt.Visible then
    FVisCols := (ClientWidth - FGutterPan.Width) div FChrW
  else
    FVisCols := ClientWidth div FChrW;
end;

procedure TGsCustomMemo.Paint;
begin
  UpdateCharBounds;

//  FBuffer.Width := ClientWidth;
//  FBuffer.Height := ClientHeight;

  if FGutterOpt.Visible and FGutterPan.ShowNum and (FGutterPan.StartNum <> FTopLine + 1) then
  begin
    FGutterPan.StartNum := FTopLine + 1;
    FGutterPan.DrawLineNums;
  end;
  with {FBuffer.}Canvas do
  begin
    Brush.Color := Self.Color;
    FillRect(Canvas.ClipRect);
    if not FGutterOpt.Visible then
    begin
      FGutterPan.Visible := False;
      FGutterPan.Align := alNone;
      FGutterPan.Left := -FGutterPan.Width - 1;
    end else
    begin
      FGutterPan.Visible := True;
      FGutterPan.Align := alLeft;
    end;
    Brush.Color := Self.Color;
  end;
  DrawVisible;
  Canvas.Draw(0, 0, FBuffer);
  if Assigned(FOnPaint) then FOnPaint(Self);
end;

{ Format visible lines and Drawing visible lines }

procedure TGsCustomMemo.MarkHyperlinks(S: string; var LGliph: TBitmap);
var
  HLS, HS: string;
  SP, HL, KeepSP: Integer;
  NextHL: Boolean;
  R: TRect;

  function FindHyperlinks(URLFind: Boolean): Boolean;
  begin
    Result := False;
    if URLFind then
      REx.Expression := RegularURL
    else
      REx.Expression := RegularEmail;
    if REx.Exec(HLS, 1) then
    begin
      SP := REx.MatchPos[0] - 1;
      HL := REx.MatchLen[0];
      Result := True;
    end;
  end;

begin
  HLS := S;
  NextHL := True;
  KeepSP := 0;
  while NextHL do
  begin
    if FindHyperlinks(True) then
    begin
      LGliph.Canvas.Font.Color := FEnvironment.Hyperlink;
      LGliph.Canvas.Font.Style := [fsUnderline];
      LGliph.Canvas.Brush.Color := FEnvironment.HyperlinkBackground;
      HS := Copy(HLS, SP + 1, HL);
      Delete(HLS, 1, SP + HL + 1);
      KeepSP := KeepSP + SP;
      R := Rect(KeepSP * FChrW, 0, (KeepSP * FChrW) + (Length(HS) * FChrW), FChrH);
      DrawText(LGliph.Canvas.Handle, PChar(HS), Length(HS), R, DT_DRAWLINE);
      // LGliph.Canvas.TextOut(KeepSP * FChrW, 0, HS);
      KeepSP := KeepSP + HL + 1;
    end else
      NextHL := False;
  end;
  HLS := S;
  NextHL := True;
  KeepSP := 0;
  while NextHL do
  begin
    if FindHyperlinks(False) then
    begin
      LGliph.Canvas.Font.Color := FEnvironment.Hyperlink;
      LGliph.Canvas.Font.Style := [fsUnderline];
      LGliph.Canvas.Brush.Color := FEnvironment.HyperlinkBackground;
      HS := Copy(HLS, SP + 1, HL);
      Delete(HLS, 1, SP + HL + 1);
      KeepSP := KeepSP + SP;
      R := Rect(KeepSP * FChrW, 0, (KeepSP * FChrW) + (Length(HS) * FChrW), FChrH);
      DrawText(LGliph.Canvas.Handle, PChar(HS), Length(HS), R, DT_DRAWLINE);
      // LGliph.Canvas.TextOut(KeepSP * FChrW, 0, HS);
      KeepSP := KeepSP + HL + 1;
    end else
      NextHL := False;
  end;
  LGliph.Canvas.Font.Style := [];
end;

procedure TGsCustomMemo.FormatLine(I: Integer; var LGliph: TBitmap);
var
  S, CorrectS, SS: string;
  TI, Si, Ei, T: Integer;
  R: TRect;
  AllowDraw: Boolean;
begin
  if FLines.Count = 0 then Exit;
  if (I < 0) or (I > Pred(FLines.Count)) then Exit;
  S := FLines[I];
  if Pos(#9, S) > 0 then
  begin
    CorrectS := '';
    for TI := 1 to Length(S) do
    begin
      if S[TI] = #9 then
      begin
        for T := 1 to FTabWidth do CorrectS := CorrectS + ' ';
      end else
        CorrectS := CorrectS + S[TI];
    end;
    S := CorrectS;
  end;

  LGliph.Width := FChrW * Length(S);
  LGliph.Height := FChrH;

  with LGliph.Canvas do
  begin

    R := Rect(0, 0, Length(S) * FChrW, FChrH);
    AllowDraw := True;
    if Assigned(FOnDrawLine) then FOnDrawLine(Self, S, I, LGliph, R, AllowDraw);
    { Draw simple text line... }
    if AllowDraw then
      DrawText(LGliph.Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
    // TextOut(0, 0, S);
    if FShowHyperlinks then MarkHyperlinks(S, LGliph);
    if FSelected then
    begin
      Font.Color := FEnvironment.TextColor;
      Brush.Color := FEnvironment.TextBackground;
      Font.Style := [];
      if (I > StartNo) and (I < EndNo) then
      begin
        Brush.Color := FEnvironment.SelectionBackground;
        Font.Color := FEnvironment.SelectionColor;
        R := Rect(0, 0, Length(S) * FChrW, FChrH);
        DrawText(LGliph.Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
        // TextOut(0, 0, S);
      end else
      begin
        Ei := EndOffs;
        Si := StartOffs;
        if (I = StartNo) and (I = EndNo) then
        begin
          Brush.Color := FEnvironment.SelectionBackground;
          Font.Color := FEnvironment.SelectionColor;
          SS := Copy(S, Si + 1, Length(S) - Si);
          if Ei > Length(S) then
            SS := Copy(S, Si + 1, Length(S) - Si)
          else
            SS := Copy(S, Si + 1, Ei - Si);
          R := Rect(Si * FChrW, 0, (Si * FChrW) + (Length(SS) * FChrW), FChrH);
          DrawText(LGliph.Canvas.Handle, PChar(SS), Length(SS), R, DT_DRAWLINE);
          // TextOut(Si * FChrW, 0, SS)
        end else
          if (I = StartNo) and (I < EndNo) then
          begin
            Brush.Color := FEnvironment.SelectionBackground;
            Font.Color := FEnvironment.SelectionColor;
            SS := Copy(S, Si + 1, Length(S) - Si);
            R := Rect(Si * FChrW, 0, (Si * FChrW) + (Length(SS) * FChrW), FChrH);
            DrawText(LGliph.Canvas.Handle, PChar(SS), Length(SS), R, DT_DRAWLINE);
            // TextOut(Si * FChrW, 0, SS);
          end else
            if (I > StartNo) and (I = EndNo) then
            begin
              Brush.Color := FEnvironment.SelectionBackground;
              Font.Color := FEnvironment.SelectionColor;
              if Ei > Length(S) then Ei := Length(S);
              SS := Copy(S, 1, Ei);
              R := Rect(0, 0, (Length(SS) * FChrW), FChrH);
              DrawText(LGliph.Canvas.Handle, PChar(SS), Length(SS), R, DT_DRAWLINE);
              // TextOut(0, 0, Copy(S, 1, Ei));
            end;
      end;
    end;
  end;
  if FRightEdge and ((LGliph.Width div FChrW) >= FRightEdgeCol) then
    with LGliph.Canvas do
    begin
      Pen.Color := FEnvironment.RightEdgeColor;
      MoveTo(FRightEdgeCol * FChrW, 0);
      LineTo(FRightEdgeCol * FChrW, FChrH);
    end;
  if Length(S) > FMaxScrollH then FMaxScrollH := Length(S);
  UpdateScrollBars;
end;

procedure TGsCustomMemo.DrawLine(CanvasSupport: TCanvas; I, Y: Integer);
var
  L, Ei, Si, GSz: Integer;
  LGliph: TBitmap;
  Rc: TRect;
begin
  if FGutterOpt.Visible then
  begin
    GSz := FGutterPan.Width;
    if FGutterPan.ShowNum and (FGutterPan.StartNum <> FTopLine + 1) then
    begin
      FGutterPan.StartNum := FTopLine + 1;
      FGutterPan.DrawLineNums;
    end;
  end else
    GSz := 0;
  LGliph := TBitmap.Create;
  with LGliph.Canvas do
  begin
    Brush.Color := FEnvironment.TextBackground;
    Font.Assign(Self.Font);
    Font.Color := FEnvironment.TextColor;
  end;
  if I < FLines.Count then
    FormatLine(I, LGliph)
  else
    LGliph.Width := 0;

  with CanvasSupport do
  begin
    Draw(-(HPos * FChrW) + GSz, Y, LGliph);
    L := LGliph.Width - (HPos * FChrW) + GSz;
    if L < GSz then L := GSz;
    Brush.Color := FEnvironment.TextBackground; { Set default background color }
    if FSelected then
    begin
      Ei := EndOffs;
      Si := StartOffs;
      if (I > StartNo) and (I < EndNo) then
      begin
        Brush.Color := FEnvironment.SelectionBackground;
        Rc := Rect(L, Y, (FMaxScrollH * FChrW) + ClientWidth - (HPos * FChrW), Y + FChrH);
        FillRect(Rc);
      end else
        if (I = StartNo) and (I < EndNo) then
        begin
          if I < FLines.Count then
            if Length(FLines[I]) - 1 < Si then
            begin
              Si := Length(FLines[I]) - 1;
              StartOffs := Si;
            end;
          if LGliph.Width < 1 then
          begin
            Color := Self.Color;
            Rc := Rect(-(HPos * FChrW) + GSz, Y, (Si * FChrW) - (HPos * FChrW) +
              GSz, Y + FChrH);
            FillRect(Rc);
            L := (Si * FChrW) - (HPos * FChrW) + GSz;
          end else
            L := LGliph.Width - (HPos * FChrW) + GSz;
          if LGliph.Width < (Si * FChrW) then
          begin
            Brush.Color := Self.Color;
            Rc := Rect(L, Y, (Si * FChrW) - (HPos * FChrW) + GSz, Y + FChrH);
            FillRect(Rc);
            L := (Si * FChrW) - (HPos * FChrW) + GSz;
          end;
          Brush.Color := FEnvironment.SelectionBackground;
          Rc := Rect(L, Y, (FMaxScrollH * FChrW) + ClientWidth - (HPos * FChrW) +
            GSz, Y + FChrH);
          FillRect(Rc);
        end else
          if (I > StartNo) and (I = EndNo) then
          begin
            Brush.Color := FEnvironment.SelectionBackground;
            if LGliph.Width < 1 then
            begin
              Rc := Rect(-(HPos * FChrW) + GSz, Y, (Ei * FChrW) - (HPos * FChrW) +
                GSz, Y + FChrH);
              FillRect(Rc);
            end else
              if LGliph.Width < (Ei * FChrW) then
              begin
                L := LGliph.Width - (HPos * FChrW) + GSz;
                Rc := Rect(L, Y, (Ei * FChrW) - (HPos * FChrW) + GSz, Y + FChrH);
                FillRect(Rc);
              end;
            L := (Ei * FChrW) - (HPos * FChrW) + GSz;
            if LGliph.Width - (HPos * FChrW) + GSz > L then
              L := LGliph.Width - (HPos * FChrW) + GSz;
            Brush.Color := Self.Color;
            Rc := Rect(L, Y, (FMaxScrollH * FChrW) + ClientWidth - (HPos * FChrW) +
              GSz, Y + FChrH);
            FillRect(Rc);
          end else
          begin
            Color := Self.Color;
            Rc := Rect(L, Y, (FMaxScrollH * FChrW) + ClientWidth - (HPos * FChrW) +
              GSz, Y + FChrH);
            FillRect(Rc);
          end;
    end else
    begin
      Brush.Color := Self.Color;
      Rc := Rect(L, Y, (FMaxScrollH * FChrW) + ClientWidth - (HPos * FChrW), Y + FChrH);
      FillRect(Rc);
    end;
  end;
  LGliph.Free;
end;

procedure TGsCustomMemo.DrawVisible;
var
  I, Y, cntVis: Integer;
begin
  Y := 0;
  cntVis := 1;
  GetSelBounds(StartNo, EndNo, StartOffs, EndOffs);
  for I := FTopLine to FTopLine + FVisLines do
  begin
    DrawLine({FBuffer.}Canvas, I, Y);
    Y := Y + FChrH;
    cntVis := cntVis + 1;
    if cntVis > FVisLines then Break;
  end;
  if FRightEdge then
  begin
    if not FGutterOpt.Visible then
    begin
      with {FBuffer.}Canvas do
      begin
        Pen.Color := FEnvironment.RightEdgeColor; ;
        MoveTo((FRightEdgeCol * FChrW) - (HPos * FChrW), 0);
        LineTo((FRightEdgeCol * FChrW) - (HPos * FChrW), ClientHeight);
      end;
    end else
      with {FBuffer.}Canvas do
      begin
        Pen.Color := FEnvironment.RightEdgeColor; ;
        MoveTo((FRightEdgeCol * FChrW) - (HPos * FChrW) + FGutterPan.Width, 0);
        LineTo((FRightEdgeCol * FChrW) - (HPos * FChrW) + FGutterPan.Width, ClientHeight);
      end;
  end;
  Canvas.Draw(0, 0, FBuffer);
end;

{ Selection properties implementation }

function TGsCustomMemo.GetSelBegin: TPoint;
begin
  Result.X := FSelStartNo;
  REsult.Y := FSelStartOffs;
end;

function TGsCustomMemo.GetSelEnd: TPoint;
begin
  Result.X := FSelEndNo;
  REsult.Y := FSelEndOffs;
end;

procedure TGsCustomMemo.SetSelBegin(Value: TPoint);
begin
  if Value.X > pred(FLines.Count) then Value.X := pred(FLines.Count);
  if Value.X < 0 then Value.X := 0;
  if Value.Y < 0 then Value.Y := 0;
  FSelStartNo := Value.X;
  FSelStartOffs := Value.Y;
end;

procedure TGsCustomMemo.SetSelEnd(Value: TPoint);
begin
  if Value.X > pred(FLines.Count) then Value.X := pred(FLines.Count);
  if Value.X < 0 then Value.X := 0;
  if Value.Y < 0 then Value.Y := 0;
  FSelEndNo := Value.X;
  FSelEndOffs := Value.Y;
end;

{ Mouse selection implementation }

procedure TGsCustomMemo.GetSelBounds(var StartNo, EndNo, StartOffs, EndOffs: Integer);
begin
  if FSelStartNo <= FSelEndNo then
  begin
    StartNo := FSelStartNo;
    EndNo := FSelEndNo;
    if not ((StartNo = EndNo) and (FSelStartOffs > FSelEndOffs)) then
    begin
      StartOffs := FSelStartOffs;
      EndOffs := FSelEndOffs;
    end else
    begin
      StartOffs := FSelEndOffs;
      EndOffs := FSelStartOffs;
    end;
  end else
  begin
    StartNo := FSelEndNo;
    EndNo := FSelStartNo;
    StartOffs := FSelEndOffs;
    EndOffs := FSelStartOffs;
  end;
end;

procedure TGsCustomMemo.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  RNo, CNo: Integer;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if not Focused then SetFocus;
  if FDropList.Visible then
    FDropList.Visible := False;
  FDropTimeCount := 0;    
  ShiftState := Shift;
  if Assigned(FOnMemoEvents) then
    case Button of
      mbRight: FOnMemoEvents(Self, Shift, k_Mouse2, 0, CaretPos, Point(X, Y));
      mbMiddle: FOnMemoEvents(Self, Shift, k_Mouse3, 0, CaretPos, Point(X, Y));
    end;
  if Button <> mbLeft then Exit;
  if FDblClick then Exit;
  if FGutterOpt.Visible and (X <= FGutterPan.Width) then Exit;
  GetRowColAtPos(X + HPos * FChrW, Y + VPos * FChrH, RNo, CNo);
  CaretPos.X := RNo;
  CaretPos.Y := CNo;
  if Assigned(FOnMemoEvents) then FOnMemoEvents(Self, Shift, k_Mouse1, 0, CaretPos, Point(X, Y));
  FSelDrag := False;
  if (RNo in [FSelStartNo..FSelEndNo]) or (RNo in [FSelEndNo..FSelStartNo]) then
  begin
    if (FSelStartNo = FSelEndNo) and ((CNo in [FSelStartOffs..FSelEndOffs]) or
      (CNo in [FSelEndOffs..FSelStartOffs])) then FSelDrag := True;
    if FSelStartNo <> FSelEndNo then
    begin
      FSelDrag := True;
      if (RNo = FSelStartNo) and (FSelStartNo < FSelEndNo) and (CNo < FSelStartOffs) then
        FSelDrag := False;
      if (RNo = FSelStartNo) and (FSelStartNo > FSelEndNo) and (CNo > FSelStartOffs) then
        FSelDrag := False;
      if (RNo = FSelEndNo) and (FSelStartNo < FSelEndNo) and (CNo > FSelStartOffs) then
        FSelDrag := False;
      if (RNo = FSelEndNo) and (FSelStartNo > FSelEndNo) and (CNo < FSelStartOffs) then
        FSelDrag := False;
    end;
  end;
  if FSelDrag then Exit;
  if not (ssShift in Shift) then
  begin
    if FSelected then
    begin
    { Erase old selection, if any... }
      FSelected := False;
      DrawVisible;
    end else
    begin
      PulseCaret := True;
      UpdateTimer;
    end;
    FSelStartNo := RNo;
    FSelEndNo := FSelStartNo;
    FSelStartOffs := CNo;
    FSelEndOffs := FSelStartOffs;
    FSelected := True;
    FSelMouseDwn := True;
    ScrollTimer.Enabled := True;
  end else
  begin
    FSelEndNo := RNo;
    FSelEndOffs := CNo;
    FSelected := True;
  end;
  DrawVisible;
  DrawCaret(CaretPos.X, CaretPos.Y, True);
end;

function TGsCustomMemo.GetHyperlinksAtPos(X, Y: Integer; var SLink: string): Boolean;
var
  HLS, S: string;
  SP, HL, RNo, CNo, LNo, verY: Integer;

  function FindHyperlinks(URLFind: Boolean): Boolean;
  begin
    Result := False;
    if URLFind then
      REx.Expression := RegularURL
    else
      REx.Expression := RegularEmail;
    if REx.Exec(HLS, 1) then
    begin
      SP := REx.MatchPos[0] - 1;
      HL := REx.MatchLen[0];
      Result := True;
    end;
  end;

  function GetGroupAtPos(X1, Y1: Integer; var XBegin: Integer): string;
  const
    ValidChars = ['a'..'z', 'A'..'Z', '0'..'9', '#', '.', '/', ':', '@', '_', '-', '&', '%', '=', '?'];
  var
    S1: string;
    I, Si, Ei: Integer;
  begin
    Result := '';
    XBegin := -1;
    Si := 0;
    Ei := 0;
    if Y1 > pred(FLines.Count) then Exit;
    S1 := FLines[Y1];
    if S1 = '' then Exit;
    if X1 > Length(S1) - 1 then X1 := Length(S1) - 1;
    if not (S1[X1 + 1] in ValidChars) then Exit;
    for I := (X1 + 1) downto 1 do
      if S1[I] in ValidChars then
        Si := I
      else
        Break;
    for I := (X1 + 1) to Length(S1) do
      if S1[I] in ValidChars then
        Ei := I + 1
      else
        Break;
    if Ei >= Si then
    begin
      Result := Copy(S1, Si, Ei - Si);
      XBegin := Si - 1;
    end;
  end;


begin
  verY := Y;
  Result := False;
  if verY < 0 then verY := 0;
  if verY > ClientHeight then verY := ClientHeight;
  GetRowColAtPos(X + HPos * FChrW, verY + VPos * FChrH, RNo, CNo);
  S := GetGroupAtPos(CNo, RNo, LNo);
  HLS := S;
  if FindHyperlinks(True) then
  begin
    Result := True;
    SLink := HLS;
  end else
  begin
    HLS := S;
    if FindHyperlinks(False) then
    begin
      Result := True;
      SLink := HLS;
    end;
  end;
end;

procedure TGsCustomMemo.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  RNo, CNo: Integer;
  SLink: string;
  AtPos: TPoint;
  CopyMode: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  ShiftState := Shift;
  if Assigned(FOnMemoEvents) then
    case Button of
      mbLeft: FOnMemoEvents(Self, Shift, k_Mouse1, 0, CaretPos, Point(X, Y));
      mbRight: FOnMemoEvents(Self, Shift, k_Mouse2, 0, CaretPos, Point(X, Y));
      mbMiddle: FOnMemoEvents(Self, Shift, k_Mouse3, 0, CaretPos, Point(X, Y));
    end;
  if Button <> mbLeft then Exit;
  if FSelDrag then
  begin
    Cursor := crIBeam;
    GetRowColAtPos(X + HPos * FChrW, Y + VPos * FChrH, AtPos.x, AtPos.y);
    CopyMode := (ssCtrl in Shift);
    DragDropSelection(AtPos, CopyMode);
    Screen.Cursor := Cursor;
    FSelDrag := False;
    DrawVisible;
    Exit;
  end;
  if FDblClick then
  begin
    FDblClick := False;
    Exit;
  end;
  FSelMouseDwn := False;
  ScrollTimer.Enabled := False;
  if FGutterOpt.Visible and (X <= FGutterPan.Width) then Exit;
  GetRowColAtPos(X + HPos * FChrW, Y + VPos * FChrH, RNo, CNo);
  if RNo <> FSelEndNo then FSelEndNo := RNo;
  if CNo <> FSelEndOffs then FSelEndOffs := CNo;
  FSelected := (FSelStartNo <> FSelEndNo) or (FSelStartOffs <> FSelEndOffs);
  if ssCtrl in Shift then
    if GetHyperlinksAtPos(X, Y, SLink) then
      if Assigned(FOnHyperlinkClick) then FOnHyperlinkClick(Self, SLink);
  if not FSelected and (FSelStartNo = 0) and (FSelStartOffs = 0) then
  begin
    FSelStartNo := 0;
    FSelEndNo := FSelStartNo;
    FSelStartOffs := 0;
    FSelEndOffs := FSelStartOffs;
    DrawVisible;
    DrawCaret(CaretPos.X, CaretPos.Y, True);
    HideCaret := True;
    PulseCaret := True;
    UpdateTimer;
  end;
end;

procedure TGsCustomMemo.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  RNo, CNo, verY: Integer;
  SLink: string;
begin
  inherited MouseMove(Shift, X, Y);
  ShiftState := Shift;
  MousePos := Point(X, Y);
  if Assigned(FOnMemoEvents) then FOnMemoEvents(Self, Shift, k_None, 0, CaretPos, MousePos);
  if FSelDrag then
  begin
    if ssCtrl in Shift then
      Cursor := crDragSelCopy
    else
      Cursor := crDragSel;
    if Screen.Cursor <> Cursor then
      Screen.Cursor := Cursor;
    Exit;
  end;
  VScrollDelta := 0;
  HScrollDelta := 0;
  if Y < 0 then VScrollDelta := -1;
  if X < 0 then HScrollDelta := -1;
  if Y > ClientHeight then VScrollDelta := 1;
  if X > ClientWidth then HScrollDelta := 1;
  if Y < -FChrH then VScrollDelta := -2;
  if Y > (ClientHeight + FChrH) then VScrollDelta := 2;
  if X < -FChrW then HScrollDelta := -2;
  if X > (ClientWidth + FChrW) then HScrollDelta := 2;
  if FSelected and FSelMouseDwn then
  begin
    XMouse := X;
    YMouse := Y;
    verY := Y;
    if verY < 0 then verY := 0;
    if verY > ClientHeight then verY := ClientHeight;
    GetRowColAtPos(X + HPos * FChrW, verY + VPos * FChrH, RNo, CNo);
    FSelEndNo := RNo;
    FSelEndOffs := CNo;
    CaretPos.x := RNo;
    CaretPos.y := CNo;
    DrawVisible;
  end;
  if FGutterOpt.Visible then
  begin
    if (X <= FGutterPan.Width) and (Cursor <> crArrow) then Cursor := crArrow;
    if (X > FGutterPan.Width) and (Cursor <> crIBeam) then Cursor := crIBeam;
  end;
  if ssCtrl in Shift then
  begin
    if GetHyperlinksAtPos(X, Y, SLink) then
    begin
      Cursor := crHand;
      Screen.Cursor := Cursor;
    end else
    begin
      Cursor := crIBeam;
      Screen.Cursor := Cursor;
      if (FHintWnd <> nil) and not FNotHideHint then
      begin
        FHintWnd.Free;
        FHintWnd := nil;
      end;
    end;
    Exit;
  end;
  Cursor := crIBeam;
  Screen.Cursor := Cursor;
  if FHintWnd <> nil then
  begin
    FHintWnd.Free;
    FHintWnd := nil;
  end;
end;

{ Drag'n Drop selection impementation }

procedure TGsCustomMemo.DragDropSelection(AtPos: TPoint; CopyMode: Boolean);
var
  SL, EL, SC, EC: Integer;
  S: String;
begin
  if FSelStartNo <= FSelEndNo then
  begin
    SL := FSelStartNo;
    EL := FSelEndNo;
    SC := FSelStartOffs;
    EC := FSelEndOffs;
  end else
  begin
    SL := FSelEndNo;
    EL := FSelStartNo;
    EC := FSelStartOffs;
    SC := FSelEndOffs;
  end;
  if (AtPos.x in [SL..EL]) then
  begin
    if (AtPos.x = SL) and (AtPos.x < EL) and (AtPos.y >= SC) then Exit;
    if (AtPos.x > SL) and (AtPos.x = EL) and (AtPos.y < EC) then Exit;
    if (AtPos.x > SL) and (AtPos.x < EL) then Exit;
    if (AtPos.x = SL) and (AtPos.x = EL) and (AtPos.y in [SC..EC]) then Exit;
  end;
  S := GetSelText;
  if S = '' then
  begin
    FSelected := False;
    Exit;
  end;
  if not CopyMode then Self.DeleteSelection;
  InsertTextAtPos(S, AtPos.y, AtPos.x);
end;

{ ... }

function GetNextWord(SLine: string; var PosX: Integer): Boolean;
const
  ValidChars = ['a'..'z', 'A'..'Z', '0'..'9', '#'];
var
  I, RetX: Integer;
  FindNext: Boolean;
begin
  Result := False;
  if PosX > Length(SLine) then Exit;
  FindNext := False;
  RetX := 0;
  for I := PosX to Length(SLine) do
  begin
    if not FindNext and not (SLine[I] in ValidChars) then
    begin
      FindNext := True;
      Continue;
    end;
    if FindNext and (SLine[I] in ValidChars) then
    begin
      RetX := I;
      Result := True;
      Break;
    end;
  end;
  if RetX < 1 then Result := False;
  PosX := RetX;
end;

{ Mouse Double-Click selection impementation }

procedure TGsCustomMemo.DblClick;
var
  S: string;
  XB, NextW: Integer;
begin
  inherited DblClick;
  ShiftState := ShiftState + [ssDouble];
  if Assigned(FOnMemoEvents) then FOnMemoEvents(Self, ShiftState, k_None, 0, CaretPos, MousePos);
  S := GetWordAtPos(CaretPos.y, CaretPos.x, XB);
  if (S <> '') and (XB > -1) then
  begin
    FSelStartNo := CaretPos.x;
    FSelEndNo := CaretPos.x;
    FSelStartOffs := XB;
    FSelEndOffs := XB + Length(S);
    CaretPos.y := FSelStartOffs;
    FSelected := True;
    DrawVisible;
    FDblClick := True;
  end else
    if CaretPos.x <= pred(FLines.Count) then
    begin
      S := FLines[CaretPos.x];
      NextW := CaretPos.y;
      if GetNextWord(S, NextW) then
      begin
        S := GetWordAtPos(NextW, CaretPos.x, XB);
        if (S <> '') and (XB > -1) then
        begin
          FSelStartNo := CaretPos.x;
          FSelEndNo := CaretPos.x;
          FSelStartOffs := XB;
          FSelEndOffs := XB + Length(S);
          CaretPos.y := FSelStartOffs;
          FSelected := True;
          DrawVisible;
          FDblClick := True;
        end;
      end;
    end;
end;

{ ... }

function TGsCustomMemo.GetWordAtPos(Col, Row: Integer; var XBegin: Integer): string;
const
  ValidChars = ['a'..'z', 'A'..'Z', '0'..'9', '#'];
var
  S: string;
  C: Char;
  I, Si, Ei, CrX: Integer;
begin
  Result := '';
  XBegin := -1;
  Si := 0;
  Ei := 0;
  if Row > pred(FLines.Count) then Exit;
  S := FLines[Row];
  if S = '' then Exit;
  if Col > Length(S) - 1 then Col := Length(S) - 1;
  if not (S[Col + 1] in ValidChars) then
  begin
    CrX := Col - 1;
    for I := CrX downto 1 do
    begin
      C := S[I + 1];
      if (C in ValidChars) then
      begin
        Col := I;
        Break;
      end;
    end;
    if Col = 0 then Exit;
  end;
  for I := (Col + 1) downto 1 do
    if S[I] in ValidChars then
      Si := I
    else
      Break;
  for I := (Col + 1) to Length(S) do
    if S[I] in ValidChars then
      Ei := I + 1
    else
      Break;
  if Ei >= Si then
  begin
    Result := Copy(S, Si, Ei - Si);
    XBegin := Si - 1;
  end;
end;

procedure TGsCustomMemo.GetRowColAtPos(X, Y: Integer; var Row, Col: Integer);
var
  Fine: Integer;
begin
  Row := Y div FChrH;
  if Row > Flines.Count then Row := FLines.Count;
  if FGutterOpt.Visible then
  begin
    if X < FGutterPan.Width then X := FGutterPan.Width;
    Col := (X - FGutterPan.Width) div FChrW;
    Fine := (X - FGutterPan.Width) mod FChrW;
  end else
  begin
    if X < 0 then X := 0;
    Col := X div FChrW;
    Fine := X mod FChrW;
  end;
  if Fine > (FChrW div 2) - 1 then Col := Col + 1;
end;

{ Wheel scrolling and Mouse selection scrolling implementation }

procedure TGsCustomMemo.ScrollOnTimer(Sender: TObject);
begin
  if VScrollDelta <> 0 then
  begin
    ScrollPos_V := ScrollPos_V + VScrollDelta;
    MouseMove([], XMouse, YMouse);
  end;
  if HScrollDelta <> 0 then
  begin
    ScrollPos_H := ScrollPos_H + HScrollDelta;
    MouseMove([], XMouse, YMouse);
  end;
end;

{ Caret position and Caret drawing impementation}

procedure TGsCustomMemo.BlinkCaret;
begin
  if Screen.Cursor = crIBeam then Screen.Cursor := crDefault; { ??? }
  if csDesigning in ComponentState then Exit;
  HideCaret := not HideCaret;
  if (FSelStartNo = FSelEndNo) and (FSelStartOffs = FSelEndOffs) then FSelected := False;
  if FSelected then
    HideCaret := False;
  DrawCaret(CaretPos.X, CaretPos.Y, HideCaret);
  if FDropTime = 0 then Exit;
  if FDropList.Visible then Exit;
  if (FLastDropPos.x = CaretPos.x) and (FLastDropPos.y = CaretPos.y) then Exit;
  FDropTimeCount := FDropTimeCount + 500;
  if FDropTimeCount >= (FDropTime * 1000) then
  begin
    FDropTimeCount := 0;
    ShowDropWindow;
  end;  
end;

procedure TGsCustomMemo.DrawCaret(X, Y: Integer; ShowCaret: Boolean);
var
  Xp, Yp: Integer;
begin
  if csDesigning in ComponentState then Exit;
  if (X < FTopLine) or (X > FTopLine + FVisLines) then Exit;
  Yp := (X - FTopLine) * FChrH;
  Xp := (Y - HPos) * FChrW;
  if FGutterOpt.Visible then Xp := Xp + FGutterPan.Width;
  if (Xp < 0) or (Xp > ClientWidth) then Exit;
  with Canvas do
  begin
    if ShowCaret then
    begin
      Pen.Mode := pmNotMerge;
      Pen.Color := Font.Color;
    end else
    begin
      if not FSelected then
        Pen.Color := Self.Color
      else
        Pen.Color := FEnvironment.SelectionBackground;
      Pen.Mode := pmCopy;
    end;
    MoveTo(Xp, Yp);
    LineTo(Xp, Yp + FChrH);
    Pen.Mode := pmCopy;
  end;
  if not FSelected then
  begin
    FSelStartNo := CaretPos.x;
    FSelStartOffs := CaretPos.y;
  end;
end;

function TGsCustomMemo.GetSelText: string;
var
  StartLine, StartPos, EndLine, EndPos, I, LineI: Integer;
  FirstPart, LastPart, SLine: string;
begin
  Result := '';
  if not FSelected then Exit;
  if not FSelected then Exit;
  if FSelStartNo > FSelEndNo then
  begin
    StartLine := FSelEndNo;
    StartPos := FSelEndOffs;
    EndLine := FSelStartNo;
    EndPos := FSelStartOffs;
  end else
    if (FSelStartNo = FSelEndNo) and (FSelEndOffs < FSelStartOffs) then
    begin
      StartLine := FSelStartNo;
      StartPos := FSelEndOffs;
      EndLine := StartLine;
      EndPos := FSelStartOffs;
    end else
    begin
      StartLine := FSelStartNo;
      StartPos := FSelStartOffs;
      EndLine := FSelEndNo;
      EndPos := FSelEndOffs;
    end;
  if StartLine > pred(FLines.Count) then Exit;
  if EndLine > pred(FLines.Count) then EndLine := pred(FLines.Count);
  SLine := FLines[StartLine];
  if StartLine < EndLine then
  begin
    FirstPart := Copy(SLine, StartPos + 1, Length(SLine) - StartPos);
    SLine := FLines[EndLine];
    if EndPos > Length(SLine) then EndPos := Length(SLine);
    LastPart := Copy(SLine, 1, EndPos);
    LineI := StartLine + 1;
    Result := FirstPart;
    for I := LineI to (EndLine - 1) do Result := Result + #13#10 + FLines[I];
    Result := Result + #13#10 + LastPart;
  end else
    Result := Copy(SLine, StartPos + 1, EndPos - StartPos);
end;

procedure TGsCustomMemo.CopyToClipboard;
var
  PC: PChar;
begin
  if not FSelected then Exit;
  PC := PChar(GetSelText);
  Clipboard.SetTextBuf(PC);
end;

procedure TGsCustomMemo.CutToClipboard;
begin
  if not FSelected then Exit;
//  FUndoList.AddChange(mwcrSelDelete, fBlockBegin, fBlockEnd, PChar(SelText));
  CopyToClipboard;
  DeleteSelection;
end;

procedure TGsCustomMemo.PasteFromClipboard;
begin
  if FSelected then DeleteSelection;
  InsertTextAtPos(Clipboard.AsText, CaretPos.y, CaretPos.x);
end;

procedure TGsCustomMemo.InsertTextAtPos(S: string; Col, Row: Integer);
var
  SLine, BufS1, BufS2, BufS: String;
  I, L: Integer;
begin
  if S = '' then Exit;
  if Row > FLines.Count then Exit;
  if Row = FLines.Count then FLines.Add('');
  SLine := FLines[Row];
  if Col > Length(SLine) then
  begin
    L := Length(SLine);
    for I := L to Col do SLine := Sline + ' ';
  end;  
  BufS1 := Copy(SLine, 1, Col);
  BufS2 := Copy(SLine, Col + 1, Length(SLine) - Col);
  SLine := BufS1 + S + BufS2;
  FSelected := True;  
  I := pos(#13#10, SLine);
  if I > 0 then
  begin
    BufS := '';
    FSelStartNo := Row;
    FSelStartOffs := Length(BufS1);
    while I > 0 do
    begin
      BufS := Copy(SLine, 1, I - 1);
      FLines.Insert(Row, BufS);
      Delete(SLine, 1, I + 1);
      I := pos(#13#10, SLine);
      CaretPos.x := Row;
      CaretPos.y := Length(BufS);
      FSelEndNo := Row;
      FSelEndOffs := CaretPos.y;
      Row := Row + 1;
    end;
    if SLine <> '' then
    begin
      FLines[Row] := SLine;
      CaretPos.x := Row;
      CaretPos.y := Length(SLine) - Length(BufS2);
      FSelEndNo := Row;
      FSelEndOffs := CaretPos.y;
    end;
    Invalidate;
  end else
  begin
    I := pos(#13, SLine);
    if I > 0 then
    begin
      BufS := '';
      FSelStartNo := Row;
      FSelStartOffs := Length(BufS1);
      while I > 0 do
      begin
        BufS := Copy(SLine, 1, I - 1);
        FLines.Insert(Row, BufS);
        Delete(SLine, 1, I);
        I := pos(#13, SLine);
        CaretPos.x := Row;
        CaretPos.y := Length(BufS);
        FSelEndNo := Row;
        FSelEndOffs := CaretPos.y;
        Row := Row + 1;
      end;
      if SLine <> '' then
      begin
        FLines[Row] := SLine;
        CaretPos.x := Row;
        CaretPos.y := Length(SLine) - Length(BufS2);
        FSelEndNo := Row;
        FSelEndOffs := CaretPos.y;
      end;
      Invalidate;
    end else
    begin
      CaretPos.x := Row;
      FLines[Row] := SLine;
      CaretPos.y := Col + Length(S);
      FSelStartNo := Row;
      FSelEndNo := Row;
      FSelStartOffs := Length(BufS1);
      FSelEndOffs := CaretPos.y;
      Invalidate;
    end;
  end;  
end;

{ Keyboard implementation }

procedure TGsCustomMemo.DeleteSelection;
var
  FirstPart, LastPart, SLine: string;
  StartLine, StartPos, EndLine, EndPos, I, DelLine: Integer;
begin
  if not FSelected then Exit;
  if FSelStartNo > FSelEndNo then
  begin
    StartLine := FSelEndNo;
    StartPos := FSelEndOffs;
    EndLine := FSelStartNo;
    EndPos := FSelStartOffs;
  end else
    if (FSelStartNo = FSelEndNo) and (FSelEndOffs < FSelStartOffs) then
    begin
      StartLine := FSelStartNo;
      StartPos := FSelEndOffs;
      EndLine := StartLine;
      EndPos := FSelStartOffs;
    end else
    begin
      StartLine := FSelStartNo;
      StartPos := FSelStartOffs;
      EndLine := FSelEndNo;
      EndPos := FSelEndOffs;
    end;
  { update Undo List }

  if StartLine > pred(FLines.Count) then Exit;
  if EndLine > pred(FLines.Count) then EndLine := pred(FLines.Count);
  SLine := FLines[StartLine];
  FirstPart := Copy(SLine, 1, StartPos);
  SLine := FLines[EndLine];
  if EndPos > Length(SLine) then EndPos := Length(SLine);
  LastPart := Copy(SLine, EndPos + 1, Length(SLine) - EndPos);
  DelLine := StartLine + 1;
  for I := DelLine to EndLine do FLines.Delete(DelLine);
  FLines[StartLine] := FirstPart + LastPart;

  CaretPos.x := StartLine;
  CaretPos.y := StartPos;
  FSelected := False;
  if StartLine = EndLine then
  begin
    I := (CaretPos.x - FTopLine) * FChrH;
    DrawLine(Canvas, CaretPos.x, I);
  end else
    DrawVisible;
  DrawCaret(CaretPos.x, CaretPos.y, True);
  UpdateScrollbars;
end;

{ Showing Special Hint Window - Text only }

procedure TGsCustomMemo.ShowTextHint(X, Y: Integer; HintMsg: string);
var
  HintPos: TPoint;
  WTxt, HTxt: Integer;
begin
  if FHintWnd <> nil then
  begin
    Exit;
    FHintWnd.Free;
    FHintWnd := nil;
  end;
  FHintWnd := THintWindow.Create(Self);
  HintPos := ClientToScreen(Point(X, Y));
  with FHintWnd do
  begin
    Color := clInfoBk;
    Font.Name := 'MS Sans Serif';
    Font.Size := 8;
    HTxt := Canvas.TextHeight(HintMsg);
    WTxt := Canvas.TextWidth(HintMsg);
    ActivateHint(Rect(HintPos.x, HintPos.y, HintPos.x + WTxt + 8, HintPos.y + HTxt + 2), HintMsg);
  end;
end;

procedure TGsCustomMemo.WMKeyDown(var Message: TWMKeyDown);
var
  GliphY, Y: Integer;
  CaretScroll: Boolean;
  SLine, AddS, SLink: string;
begin
  CaretScroll := False;
  FDropTimeCount := 0;
  FLastDropPos.x := -1;
  FLastDropPos.y := -1;
  with Message do
  begin
    ShiftState := KeyDataToShiftState(KeyData);
    if Assigned(FOnMemoEvents) then
      FOnMemoEvents(Self, ShiftState, k_Down, CharCode, CaretPos, MousePos);
    if not FSelDrag and (ssCtrl in ShiftState) then
    begin
      if GetHyperlinksAtPos(MousePos.x, MousePos.y, SLink) then
      begin
        Cursor := crHand;
        Screen.Cursor := Cursor;
        if FHintWnd <> nil then Exit;
        AddS := LowerCase(Copy(SLink, Length(SLink) - 3, 4));
        if (AddS = '.jpg') or (AddS = '.gif') or (AddS = '.jpeg') then
          ShowTextHint(MousePos.x - 2, MousePos.y + 20, S_IMAGE_HINT + SLink)
        else
          ShowTextHint(MousePos.x - 2, MousePos.y + 20, S_LINK_HINT + SLink);
      end else
      begin
        Cursor := crIBeam;
        Screen.Cursor := Cursor;
      end;
    end;
    if FSelDrag and (ssCtrl in ShiftState) then
    begin
      Cursor := crDragSelCopy;
      Screen.Cursor := Cursor;
      if FHintWnd <> nil then
      begin
        FHintWnd.Free;
        FHintWnd := nil;
      end;
    end;
    case CharCode of
      VK_LEFT, VK_RIGHT, VK_DOWN, VK_UP, VK_HOME, VK_END, VK_PRIOR, VK_NEXT:
        begin
          KeyboardCaretNav(ShiftState, CharCode);
          CaretScroll := True;
        end;
      VK_DELETE:
        begin
          if FSelected then
          begin
            DeleteSelection;
            Exit;
          end;
          if CaretPos.x > pred(FLines.Count) then Exit;
          SLine := FLines[CaretPos.x];
          if Length(SLine) >= CaretPos.y + 1 then
          begin
            Y := CaretPos.y + 1;
            Delete(SLine, Y, 1);
            FLines[CaretPos.x] := SLine;
            GliphY := (CaretPos.x - FTopLine) * FChrH;
            DrawLine(Canvas, CaretPos.x, GliphY);
          end else
          begin
            if CaretPos.x + 1 > pred(FLines.Count) then Exit;
            AddS := FLines[CaretPos.x + 1];
            FLines[CaretPos.x] := SLine + AddS;
            FLines.Delete(CaretPos.x + 1);
            DrawVisible;
          end;
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;

    end;
  end;
  if CaretScroll then
  begin
    if CaretPos.y > HPos + FVisCols then ScrollPos_H := CaretPos.y - FVisCols;
    if CaretPos.y < HPos then ScrollPos_H := CaretPos.y;
    if CaretPos.x < FTopLine then ScrollPos_V := CaretPos.x;
    if CaretPos.x > FTopLine + FVisLines - 2 then ScrollPos_V := CaretPos.x - FVisLines + 2;
  end;
  inherited;
end;

procedure TGsCustomMemo.WMKeyUp(var Message: TWMKeyUp);
begin
  with Message do
  begin
    ShiftState := KeyDataToShiftState(KeyData);
    if Assigned(FOnMemoEvents) then
      FOnMemoEvents(Self, ShiftState, k_Up, CharCode, CaretPos, MousePos);
    if FSelDrag and not (ssCtrl in ShiftState) then
    begin
      Cursor := crDragSel;
      Screen.Cursor := Cursor;
    end else
    begin
      Cursor := crIBeam;
      Screen.Cursor := Cursor;
    end;
    if FHintWnd <> nil then
    begin
      FHintWnd.Free;
      FHintWnd := nil;
      FNotHideHint := False;
    end;
  end;
  inherited;
end;

{ KeyboardCaretNav - this procedure is used to set caret position
  on keyboard navigation and to set selection if Shift key is pressed. }

procedure TGsCustomMemo.KeyboardCaretNav(ShiftState: TShiftState; Direction: Word);
var
  GliphY, SaveXCaret: Integer;

  procedure CtrlKeyLeftKey;
  var
    S: string;
    XB: Integer;
  begin
    S := GetWordAtPos(CaretPos.y, CaretPos.x, XB);
    if (S <> '') and (XB > -1) then
    begin
      if FSelected then FSelEndOffs := XB;
      CaretPos.y := XB;
    end else
    begin
      if FSelected then FSelEndOffs := 0;
      CaretPos.y := 0;
    end;
  end;

  procedure CtrlKeyRightKey;
  var
    S: string;
    I: Integer;
    NotFindIt: Boolean;
  begin
    if CaretPos.x <= pred(FLines.Count) then
    begin
      NotFindIt := True;
      while NotFindIt do
      begin
        S := FLines[CaretPos.x];
        I := CaretPos.y;
        if GetNextWord(S, I) then
        begin
          CaretPos.y := I - 1;
          NotFindIt := False;
        end else
        begin
          CaretPos.x := CaretPos.x + 1;
          CaretPos.y := 1;
        end;
        if CaretPos.x > pred(FLines.Count) then
        begin
          NotFindIt := False;
          CaretPos.y := 0;
        end;
      end;
    end else
      CaretPos.y := 0;
  end;

begin
  case Direction of
    VK_LEFT: { Left Arrow key is pressed. }
      begin
        CaretPos.y := CaretPos.y - 1;
        if CaretPos.y < 0 then
        begin
          if CaretPos.x > 0 then
          begin
            if CaretPos.x <= pred(FLines.Count) then
            begin
              GliphY := (CaretPos.x - FTopLine) * FChrH;
              DrawLine(Canvas, CaretPos.x, GliphY);
              DrawCaret(CaretPos.x, CaretPos.y + 1, False);
              if (ssCtrl in ShiftState) and (CaretPos.x > 0) then
              begin
                CaretPos.x := CaretPos.x - 1;
                CaretPos.y := Length(FLines[CaretPos.x]);
                if FSelected then
                begin
                  FSelEndNo := CaretPos.x;
                  FSelEndOffs := CaretPos.y;
                  DrawVisible;
                end;
                Exit;
              end;
            end;
            CaretPos.x := CaretPos.x - 1;
            CaretPos.y := Length(FLines[CaretPos.x]);
            if not FSelected then
            begin
              GliphY := (CaretPos.x - FTopLine) * FChrH;
              DrawLine(Canvas, CaretPos.x, GliphY);
            end else
              DrawVisible;
          end else
          begin
            CaretPos.y := 0;
            GliphY := (CaretPos.x - FTopLine) * FChrH;
            if not FSelected then
              DrawLine(Canvas, CaretPos.x, GliphY)
            else
              DrawVisible;
          end;
        end else
        begin
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;
        if ssShift in ShiftState then
        begin
          if not FSelected then
          begin
            if CaretPos.x <= pred(FLines.Count) then
              if CaretPos.y > Length(FLines[CaretPos.x]) then
                CaretPos.y := Length(FLines[CaretPos.x]) - 1;
            FSelected := True;
            FSelStartNo := CaretPos.x;
            FSelStartOffs := CaretPos.y + 1;
            FSelEndNo := CaretPos.x;
            if ssCtrl in ShiftState then
              CtrlKeyLeftKey
            else
              FSelEndOffs := CaretPos.y;
          end else
          begin
            FSelEndNo := CaretPos.x;
            if ssCtrl in ShiftState then
              CtrlKeyLeftKey
            else
              FSelEndOffs := CaretPos.y;
            if FSelEndNo <= pred(FLines.Count) then
            begin
              if FSelEndOffs > Length(FLines[FSelEndNo]) then
              begin
                FSelEndOffs := Length(FLines[FSelEndNo]) - 1;
                CaretPos.y := FSelEndOffs;
              end;
            end else
            begin
              FSelEndOffs := 0;
              CaretPos.y := 0;
            end;
          end;
          FSelected := (FSelStartNo <> FSelEndNo) or (FSelStartOffs <> FSelEndOffs);
          DrawVisible;
          Exit;
        end;
        if FSelected then
        begin
          FSelected := False;
          DrawVisible;
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;
        if ssCtrl in ShiftState then
        begin
          CtrlKeyLeftKey;
          DrawVisible;
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;
        FSelStartNo := CaretPos.x;
        FSelStartOffs := CaretPos.y;
      end;
    VK_RIGHT: { Right Arrow key is pressed. }
      begin
        CaretPos.y := CaretPos.y + 1;
        if CaretPos.y > FMaxScrollH then
        begin
          FMaxScrollH := FMaxScrollH + 2;
          UpdateScrollBars;
        end;
        if CaretPos.x <= pred(FLines.Count) then
        begin
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
        end;
        if ssShift in ShiftState then
        begin
          if not FSelected then
          begin
            FSelected := True;
            FSelStartNo := CaretPos.x;
            FSelStartOffs := CaretPos.y - 1;
            if ssCtrl in ShiftState then CtrlKeyRightKey;
            FSelEndNo := CaretPos.x;
            FSelEndOffs := CaretPos.y;
          end else
          begin
            if ssCtrl in ShiftState then CtrlKeyRightKey;
            FSelEndNo := CaretPos.x;
            FSelEndOffs := CaretPos.y;
          end;
          FSelected := (FSelStartNo <> FSelEndNo) or (FSelStartOffs <> FSelEndOffs);
          DrawVisible;
          Exit;
        end;
        if FSelected then
        begin
          FSelected := False;
          DrawVisible;
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;
        if ssCtrl in ShiftState then
        begin
          CtrlKeyRightKey;
          DrawVisible;
        end;
        DrawCaret(CaretPos.x, CaretPos.y, True);
        FSelStartNo := CaretPos.x;
        FSelStartOffs := CaretPos.y;
      end;
    VK_UP: { Up Arrow key is pressed. }
      begin
        if CaretPos.x = 0 then Exit;
        if not (ssShift in ShiftState) and not (ssCtrl in ShiftState) then
        begin
          CaretPos.x := CaretPos.x - 1;
          if FSelected then
          begin
            FSelected := False;
            DrawVisible;
            DrawCaret(CaretPos.x, CaretPos.y, True);
            Exit;
          end;
          FSelStartNo := CaretPos.x;
          GliphY := (CaretPos.x - FTopLine + 1) * FChrH;
          DrawLine(Canvas, CaretPos.x + 1, GliphY);
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
          DrawCaret(CaretPos.x, CaretPos.y, True);
          Exit;
        end;
        if (ssCtrl in ShiftState) and not (ssShift in ShiftState) then
        begin
          CaretPos.x := CaretPos.x - 1;
          Perform(WM_VSCROLL, SB_LINEUP, 0);
          FSelStartNo := CaretPos.x;
          if VPos = 0 then
          begin
            GliphY := (CaretPos.x - FTopLine + 1) * FChrH;
            DrawLine(Canvas, CaretPos.x + 1, GliphY);
            GliphY := (CaretPos.x - FTopLine) * FChrH;
            DrawLine(Canvas, CaretPos.x, GliphY);
          end;
          DrawCaret(CaretPos.x, CaretPos.y, True);
          Exit;
        end;
        if not (ssCtrl in ShiftState) and (ssShift in ShiftState) then
        begin
          CaretPos.x := CaretPos.x - 1;
          if not FSelected then
          begin
            FSelStartNo := CaretPos.x + 1;
            FSelStartOffs := CaretPos.y;
            FSelEndNo := CaretPos.x;
            FSelEndOffs := CaretPos.y;
            FSelected := True;
          end else
          begin
            FSelEndNo := CaretPos.x;
            FSelEndOffs := CaretPos.y;
            FSelected := (FSelStartNo <> FSelEndNo) or (FSelStartOffs <> FSelEndOffs);
          end;
          DrawVisible;
        end;
      end;
    VK_DOWN: { Down Arrow key is pressed. }
      begin
        if CaretPos.x > FLines.Count then Exit;
        if not (ssShift in ShiftState) and not (ssCtrl in ShiftState) then
        begin
          CaretPos.x := CaretPos.x + 1;
          if FSelected then
          begin
            FSelected := False;
            DrawVisible;
            DrawCaret(CaretPos.x, CaretPos.y, True);
            Exit;
          end;
          FSelStartNo := CaretPos.x;
          GliphY := (CaretPos.x - FTopLine - 1) * FChrH;
          DrawLine(Canvas, CaretPos.x - 1, GliphY);
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
          DrawCaret(CaretPos.x, CaretPos.y, True);
          Exit;
        end;
        if (ssCtrl in ShiftState) and not (ssShift in ShiftState) then
        begin
          CaretPos.x := CaretPos.x + 1;
          Perform(WM_VSCROLL, SB_LINEDOWN, 0);
          FSelStartNo := CaretPos.x;
          if VPos > FLines.Count - FVisLines then
          begin
            GliphY := (CaretPos.x - FTopLine - 1) * FChrH;
            DrawLine(Canvas, CaretPos.x - 1, GliphY);
            GliphY := (CaretPos.x - FTopLine) * FChrH;
            DrawLine(Canvas, CaretPos.x, GliphY);
          end;
          DrawCaret(CaretPos.x, CaretPos.y, True);
          Exit;
        end;
        if not (ssCtrl in ShiftState) and (ssShift in ShiftState) then
        begin
          CaretPos.x := CaretPos.x + 1;
          if not FSelected then
          begin
            FSelStartNo := CaretPos.x - 1;
            FSelStartOffs := CaretPos.y;
            FSelEndNo := CaretPos.x;
            FSelEndOffs := CaretPos.y;
            FSelected := True;
          end else
          begin
            FSelEndNo := CaretPos.x;
            FSelEndOffs := CaretPos.y;
            FSelected := (FSelStartNo <> FSelEndNo) or (FSelStartOffs <> FSelEndOffs);
          end;
          DrawVisible;
        end;
      end;
    VK_HOME: { Home key is pressed. }
      begin
        if not (ssCtrl in ShiftState) and not (ssShift in ShiftState) then
        begin
          CaretPos.y := 0;
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;
        if ssCtrl in ShiftState then
        begin
          if ssShift in ShiftState then
          begin
            if not FSelected then
            begin
              FSelStartNo := CaretPos.x;
              FSelStartOffs := CaretPos.y;
              FSelected := True;
            end;
            CaretPos.x := 0;
            CaretPos.y := 0;
            FSelEndNo := 0;
            FSelEndOffs := 0;
            DrawVisible;
          end else
          begin
            DrawCaret(CaretPos.x, CaretPos.y, False);
            CaretPos.x := 0;
            CaretPos.y := 0;
          end;
          Exit;
        end;
        if ssShift in ShiftState then
        begin
          DrawCaret(CaretPos.x, CaretPos.y, False);
          if not FSelected then
          begin
            FSelStartNo := CaretPos.x;
            FSelStartOffs := CaretPos.y;
            FSelected := True;
          end;
          CaretPos.y := 0;
          FSelEndNo := CaretPos.x;
          FSelEndOffs := 0;
          if FSelEndNo = FSelStartNo then FSelected := (FSelStartOffs <> FSelEndOffs);
          DrawVisible;
        end;
      end;
    VK_END: { End key is pressed. }
      begin
        if not (ssCtrl in ShiftState) and not (ssShift in ShiftState) then
        begin
          if CaretPos.x <= pred(FLines.Count) then
            CaretPos.y := Length(FLines[CaretPos.x])
          else
            CaretPos.y := 0;
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
          DrawCaret(CaretPos.x, CaretPos.y, True);
        end;
        if ssCtrl in ShiftState then
        begin
          if ssShift in ShiftState then
          begin
            if not FSelected then
            begin
              FSelStartNo := CaretPos.x;
              FSelStartOffs := CaretPos.y;
              FSelected := True;
            end;
            CaretPos.x := pred(FLines.Count);
            CaretPos.y := Length(FLines[CaretPos.x]);
            FSelEndNo := pred(FLines.Count);
            FSelEndOffs := Length(FLines[CaretPos.x]);
            DrawVisible;
          end else
          begin
            DrawCaret(CaretPos.x, CaretPos.y, False);
            CaretPos.x := pred(FLines.Count);
            CaretPos.y := Length(FLines[CaretPos.x]);
          end;
          Exit;
        end;
        if ssShift in ShiftState then
        begin
          DrawCaret(CaretPos.x, CaretPos.y, False);
          if not FSelected then
          begin
            FSelStartNo := CaretPos.x;
            if CaretPos.x <= pred(FLines.Count) then
              if CaretPos.y > Length(FLines[CaretPos.x]) then
                CaretPos.y := Length(FLines[CaretPos.x]);
            FSelStartOffs := CaretPos.y;
            FSelected := True;
          end;
          if CaretPos.x <= pred(FLines.Count) then
            CaretPos.y := Length(FLines[CaretPos.x])
          else
            CaretPos.y := 0;
          FSelEndNo := CaretPos.x;
          FSelEndOffs := CaretPos.y;
          if FSelEndNo = FSelStartNo then FSelected := (FSelStartOffs <> FSelEndOffs);
          DrawVisible;
        end;
      end;
    VK_PRIOR, VK_NEXT: { Page Up key or Page Down key is pressed. }
      begin
        if not FSelected then
        begin
          FSelStartNo := CaretPos.x;
          FSelStartOffs := CaretPos.y;
        end;
        SaveXCaret := CaretPos.x - FTopLine;
        if Direction = VK_PRIOR then
        begin
          if VPos = 0 then
          begin
            DrawCaret(CaretPos.x, CaretPos.y, False);
            CaretPos.x := 0;
            CaretPos.y := 0;
          end else
          begin
            Perform(WM_VSCROLL, SB_PAGEUP, 0);
            CaretPos.x := FTopLine + SaveXCaret;
          end;
        end else
        begin
          if VPos > FLines.Count - FVisLines then
          begin
            DrawCaret(CaretPos.x, CaretPos.y, False);
            CaretPos.x := pred(FLines.Count);
            CaretPos.y := Length(FLines[CaretPos.x]);
          end else
          begin
            Perform(WM_VSCROLL, SB_PAGEDOWN, 0);
            CaretPos.x := FTopLine + SaveXCaret;
          end;
        end;
        DrawCaret(CaretPos.x, CaretPos.y, True);
        if ssShift in ShiftState then
        begin
          FSelEndNo := CaretPos.x;
          FSelEndOffs := CaretPos.y;
          if not FSelected then FSelected := True;
          DrawVisible;
        end;
      end;
  end;

end;

procedure TGsCustomMemo.KeyPress(var Key: Char);
var
  SLine, AddS: string;
  Fill, Y, GliphY, T: Integer;
  KeyOrd: Word;
begin
  inherited KeyPress(Key);
  KeyOrd := ord(Key);
  if CaretPos.x > pred(FLines.Count) then FLines.Add('');
  if CaretPos.x > pred(FLines.Count) then Exit; { ??? }
  if (KeyOrd = VK_ESCAPE) or (ssCtrl in ShiftState) then Exit;
  if FSelected then
  begin
    DeleteSelection;
    if KeyOrd = 8 then Exit;
  end;
  SLine := FLines[CaretPos.x];
  case KeyOrd of
    8: { Backspace key }
      begin
        if Length(SLine) >= CaretPos.y then
          Y := CaretPos.y
        else
        begin
          Y := Length(SLine);
          CaretPos.y := Y;
        end;
        Delete(SLine, Y, 1);
        FLines[CaretPos.x] := SLine;
        CaretPos.y := CaretPos.y - 1;
        if CaretPos.y < 0 then
        begin
          if CaretPos.x > 0 then
          begin
            AddS := FLines[CaretPos.x];
            FLines.Delete(CaretPos.x);
            CaretPos.x := CaretPos.x - 1;
            CaretPos.y := Length(FLines[CaretPos.x]);
            if AddS <> '' then
              FLines[CaretPos.x] := FLines[CaretPos.x] + AddS;
            DrawVisible;
          end else
          begin
            CaretPos.y := 0;
            GliphY := (CaretPos.x - FTopLine) * FChrH;
            DrawLine(Canvas, CaretPos.x, GliphY);
          end;
        end else
        begin
          GliphY := (CaretPos.x - FTopLine) * FChrH;
          DrawLine(Canvas, CaretPos.x, GliphY);
        end;
        FSelStartNo := CaretPos.x;
        FSelStartOffs := CaretPos.y;
      end;
    9: { Tab key }
      begin
        { FIX-ME !!! }
// FIX-ME Undo update...
        if Length(SLine) < CaretPos.y + 1 then
          for Fill := Length(SLine) to CaretPos.y + 1 do SLine := SLine + ' ';
        if (CaretPos.x > 0) and (CaretPos.x <= pred(FLines.Count)) then
        begin
          Y := CaretPos.y + 1;
          AddS := '';
          if GetNextWord(FLines[CaretPos.x - 1], Y) then
          begin
            for Fill := 1 to Y - CaretPos.y - 1 do AddS := AddS + ' ';
          end else
            for T := 1 to FTabWidth do AddS := AddS + ' ';
        end else
          for T := 1 to FTabWidth do AddS := AddS + ' ';
        Insert(AddS, SLine, CaretPos.y + 1);
        FLines[CaretPos.x] := SLine;
        CaretPos.y := CaretPos.y + Length(AddS);
        GliphY := (CaretPos.x - FTopLine) * FChrH;
        DrawLine(Canvas, CaretPos.x, GliphY);
        FSelStartNo := CaretPos.x;
        FSelStartOffs := CaretPos.y;
      end;
    13: { Return key }
      begin
// FIX-ME Undo update...
        AddS := '';
        if Length(SLine) > CaretPos.y then
        begin
          AddS := Copy(SLine, CaretPos.y + 1, Length(SLine) - CaretPos.y + 1);
          Delete(SLine, CaretPos.y + 1, Length(SLine) - CaretPos.y);
          FLines[CaretPos.x] := SLine;
        end;
        if CaretPos.x = pred(FLines.Count) then
          FLines.Add(AddS)
        else
          if CaretPos.x < pred(FLines.Count) then
            FLines.Insert(CaretPos.x + 1, AddS)
          else
            if CaretPos.x > FLines.Count then
              FLines.Add(''); { ??? }
        CaretPos.x := CaretPos.x + 1;
        CaretPos.y := 0;
        DrawVisible;
        FSelStartNo := CaretPos.x;
        FSelStartOffs := CaretPos.y;
      end;
  else
    begin
// FIX-ME Undo update...
      if KeyOrd < 32 then Exit;
      if Length(SLine) < CaretPos.y + 1 then
        for Fill := Length(SLine) to CaretPos.y + 1 do SLine := SLine + ' ';
      AddS := Key;
      Insert(AddS, SLine, CaretPos.y + 1);
      FLines[CaretPos.x] := SLine;
      CaretPos.y := CaretPos.y + 1;
      GliphY := (CaretPos.x - FTopLine) * FChrH;
      DrawLine(Canvas, CaretPos.x, GliphY);
      FSelStartNo := CaretPos.x;
      FSelStartOffs := CaretPos.y;
    end;
  end;

  if CaretPos.y > HPos + FVisCols then
    ScrollPos_H := CaretPos.y - FVisCols;
  if CaretPos.y < HPos then
    ScrollPos_H := CaretPos.y;
  if CaretPos.x < FTopLine then
    ScrollPos_V := CaretPos.x;
  if CaretPos.x > FTopLine + FVisLines - 2 then
    ScrollPos_V := CaretPos.x - FVisLines + 2;
  DrawCaret(CaretPos.x, CaretPos.y, True);
end;

function TGsCustomMemo.FindReplaceProc(TextToFind: String; FindOptions: TGsFindOptions;
  Backward, ReplaceMode: Boolean; var ReplaceText: Boolean): Boolean;
var
  SrcBegin, SrcEnd, I, WordPos, ScrollX, ScrollY, Fill: Integer;
  SLine, SrcWord: String;
  FindPos: TPoint;
  AllowScroll, ContinueSrc: Boolean;
begin
  Result := False;
  if foEntireScope in FindOptions then
  begin
    SrcBegin := 0;
    SrcEnd := pred(FLines.Count);
  end else
  begin
    SrcBegin := CaretPos_V;
    if Backward then
      SrcEnd := 0
    else
      SrcEnd := pred(FLines.Count);
  end;
  if not (foMatchCase in FindOptions) then
    SrcWord := UpperCase(TextToFind)
  else
    SrcWord := TextToFind;
  if SrcBegin <= SrcEnd then
  begin
    for I := SrcBegin to SrcEnd do
    begin
      SLine := FLines[I];
      if not (foMatchCase in FindOptions) then SLine := UpperCase(SLine);
      FindPos.x := 0;
      WordPos := Pos(SrcWord, SLine);
      while WordPos > 0 do
      begin
        if (I = CaretPos.x) and (WordPos < CaretPos.y) then
        begin
          for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
          FindPos.x := FindPos.x + WordPos;
          WordPos := Pos(SrcWord, SLine);
          Continue;
        end;
        FindPos.x := WordPos;
        FindPos.y := I;
        AllowScroll := True;
        ContinueSrc := False;
        if foWholeWords in FindOptions then
        begin
          if WordPos > 1 then
            if (SLine[WordPos - 1] in ['a'..'z', 'A'..'Z']) then
            begin
              for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
              FindPos.x := FindPos.x + WordPos;
              WordPos := Pos(SrcWord, SLine);
              Continue;
            end;
          if WordPos + Length(SrcWord) <= Length(SLine) then
            if (SLine[WordPos + Length(SrcWord)] in ['a'..'z', 'A'..'Z']) then
            begin
              for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
              FindPos.x := FindPos.x + WordPos;
              WordPos := Pos(SrcWord, SLine);
              Continue;
            end;
        end;
        FSelStartNo := I;
        FSelEndNo := I;
        Self.FSelStartOffs := FindPos.x - 1;
        FSelEndOffs := FindPos.x + Length(SrcWord) - 1;
        FSelected := True;
        CaretPos.x := I;
        CaretPos.y := FindPos.x + Length(SrcWord) - 1;
        if AllowScroll then
        begin
          ScrollX := 0;
          ScrollY := (FTopLine * FChrH);
          if ((FindPos.x + Length(SrcWord)) * FChrW) - FChrW > ClientWidth then
            ScrollX := (FindPos.x * FChrW) - 2 * FChrW;
          if I > FTopLine + FVisLines - 2 then
            ScrollY := I * FChrH;
          ScrollTo(ScrollX, ScrollY);
          Invalidate;
        end else
          Invalidate;
        Result := True;
        if ReplaceMode then
        begin
          if Assigned(FOnReplaceText) then
            FOnReplaceText(Self, FindPos, AllowScroll, ReplaceText);
          RepPos := FindPos;
        end else
          if Assigned(FOnFindText) then
            FOnFindText(Self, FindPos, AllowScroll);
        for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
        WordPos := Pos(SrcWord, SLine);
        if not ContinueSrc then Exit;
      end;
    end;
  end else
  begin
    for I := SrcBegin downto SrcEnd do
    begin
      SLine := FLines[I];
      if not (foMatchCase in FindOptions) then SLine := UpperCase(SLine);
      FindPos.x := 0;
      WordPos := Pos(SrcWord, SLine);
      while WordPos > 0 do
      begin
        if (I = CaretPos.x) and (WordPos < CaretPos.y) then
        begin
          for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
          FindPos.x := FindPos.x + WordPos;
          WordPos := Pos(SrcWord, SLine);
          Continue;
        end;
        FindPos.x := WordPos;
        FindPos.y := I;
        AllowScroll := True;
        ContinueSrc := False;
        if foWholeWords in FindOptions then
        begin
          if WordPos > 1 then
            if (SLine[WordPos - 1] in ['a'..'z', 'A'..'Z']) then
            begin
              for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
              FindPos.x := FindPos.x + WordPos;
              WordPos := Pos(SrcWord, SLine);
              Continue;
            end;
          if WordPos + Length(SrcWord) <= Length(SLine) then
            if (SLine[WordPos + Length(SrcWord)] in ['a'..'z', 'A'..'Z']) then
            begin
              for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
              FindPos.x := FindPos.x + WordPos;
              WordPos := Pos(SrcWord, SLine);
              Continue;
            end;
        end;
        FSelStartNo := I;
        FSelEndNo := I;
        Self.FSelStartOffs := FindPos.x - 1;
        FSelEndOffs := FindPos.x + Length(SrcWord) - 1;
        FSelected := True;
        CaretPos.x := I;
        CaretPos.y := FindPos.x + Length(SrcWord) - 1;
        if AllowScroll then
        begin
          ScrollX := 0;
          ScrollY := (FTopLine * FChrH);
          if ((FindPos.x + Length(SrcWord)) * FChrW) - FChrW > ClientWidth then
            ScrollX := (FindPos.x * FChrW) - 2 * FChrW;
          if I > FTopLine + FVisLines - 2 then
            ScrollY := I * FChrH;
          ScrollTo(ScrollX, ScrollY);
          Invalidate;
        end else
          Invalidate;
        Result := True;
        if ReplaceMode then
        begin
          if Assigned(FOnReplaceText) then
            FOnReplaceText(Self, FindPos, AllowScroll, ReplaceText);
          RepPos := FindPos;
        end else
          if Assigned(FOnFindText) then
            FOnFindText(Self, FindPos, AllowScroll);
        for Fill := WordPos to WordPos + Length(SrcWord) do SLine[Fill] := '*';
        WordPos := Pos(SrcWord, SLine);
        if not ContinueSrc then Exit;
      end;
    end;
  end;
end;

procedure TGsCustomMemo.FindText(TextToFind: String; FindOptions: TGsFindOptions;
  Backward: Boolean);
var
  Rep, SrcRes: Boolean;
begin
  SrcRes := FindReplaceProc(TextToFind, FindOptions, Backward, False, Rep);
  if Assigned(FOnSerachEnd) then FOnSerachEnd(Self, SrcRes, False);
end;

procedure TGsCustomMemo.ReplaceText(TextToFind, TextToReplace: String; FindOptions: TGsFindOptions;
  Backward: Boolean);
var
  Rep, SrcRes: Boolean;
begin
  Rep := True;
  SrcRes := FindReplaceProc(TextToFind, FindOptions, Backward, True, Rep);
  if SrcRes and Rep then
    begin
      DeleteSelection;
      Self.InsertTextAtPos(TextToReplace, CaretPos.y, CaretPos.X);
    end;
  if Assigned(FOnSerachEnd) then FOnSerachEnd(Self, SrcRes, True);
end;

procedure TGsCustomMemo.Clear;
begin
  CaretPos.x := 0;
  CaretPos.y := 0;
  ScrollTo(0, 0);
  Self.FSelStartNo := 0;
  FSelStartOffs := 0;
  FSelEndNo := 0;
  FSelEndOffs := 0;
  FLines.Clear;
  FSelected := False;
  Invalidate;
end;

procedure TGsCustomMemo.SaveToFile(const FileName: string);
var
  BuffList: TStringList;
  SLine: string;
  I, P: Integer;
  Replace: Boolean;
begin
  BuffList := TStringList.Create;
  BuffList.Assign(FLines);
  for I := 0 to pred(BuffList.Count) do
  begin
    SLine := BuffList[I];
    P := Length(SLine);
    Replace := (P > 0) and (SLine <> '');
    if Replace then
    begin
      while (SLine[P] = ' ') do
      begin
        Delete(SLine, P, 1);
        P := Length(SLine);
      end;
      BuffList[I] := SLine;
    end;
  end;
  BuffList.SaveToFile(FileName);
  BuffList.Free;
end;

procedure TGsCustomMemo.LoadFromFile(const FileName: string);
begin
  if not FileExists(FileName) then Exit;
  Clear;
  FLines.LoadFromFile(FileName);
  Invalidate;
end;

{ Undo & Redo Implementation }

end.

