unit GsMemoHTMLCode;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  GsMemo;

type
  TDropItem = class(TObject)
  protected
    FList: TStringList;
    FText: String;
    procedure SetList(Value: TStringList);
  public
    constructor Create;
    destructor Destroy; override;
    property SubItems: TStringList read FList write SetList;
    property Text: String read FText write FText;
  end;

  TGsMemoHTMLCode = class(TComponent)
  private
    FCxMemo: TGsMemo;
    FCharCount, FLastPos, FStartPos: Integer;
    FInTag, FInValue, FFirstTag, FTagNotClosed: Boolean;
    FCodeHighlight: Boolean;
    FTagForeColor: TColor;
    FTagBackColor: TColor;
    FTagPropForeColor: TColor;
    FTagPropBackColor: TColor;
    FValueForeColor: TColor;
    FValueBackColor: TColor;
    FSpecTagsForeColor: TColor;
    FSpecTagsBackColor: TColor;
    FBRTagForeColor: TColor;
    FBRTagBackColor: TColor;
    FUseSpecTagsHighlight, FUseBRTagHighlight: Boolean;
    FTagsList: TStringList;

    procedure SetGsMemo(Value: TGsMemo);
    procedure SetCodeHighlight(Value: Boolean);
    procedure SetTagForeColor(Value: TColor);
    procedure SetTagBackColor(Value: TColor);
    procedure SetTagPropForeColor(Value: TColor);
    procedure SetTagPropBackColor(Value: TColor);
    procedure SetValueForeColor(Value: TColor);
    procedure SetValueBackColor(Value: TColor);
    procedure SSpecTagsForeColor(Value: TColor);
    procedure SSpecTagsBackColor(Value: TColor);
    procedure SBRTagForeColor(Value: TColor);
    procedure SBRTagBackColor(Value: TColor);
    procedure SetUseBRTagHighlight(Value: Boolean);
    procedure SetUseSpecTagsHighlight(Value: Boolean);
  protected
    procedure ShowColorHint(X, Y: Integer; HintColor: TColor);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure MemoEvents(Sender: TObject; Shift: TShiftState; KeyEvent: TGsKeyEvent;
      Key: Word; CaretPos, MousePos: TPoint);
    procedure MemoDrawLine(Sender: TObject; LineText: String; LineIndxe: Integer;
      LineGlyph: TBitmap; TextRect: TRect; var AllowSelfDraw: Boolean);
    procedure ParseLine(LineText: string; Canvas: TCanvas); dynamic;
    procedure DropListDown(Sender: TObject);
    procedure UpdateDropListForTag(TagStr: String);
    procedure UpdateDropList_AllTags;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ClearDropItems;
    function AddDropItem(DropItem: TDropItem): Integer;
    procedure DeleteDropItem(DropItemIndex: Integer);
    procedure GetDropItemsFromDTDFile(const FileName: String);
  published
    property CodeHighlight: Boolean read FCodeHighlight write SetCodeHighlight;
    property CxMemo: TGsMemo read FCxMemo write SetGsMemo;
    property TagForeColor: TColor read FTagForeColor write SetTagForeColor;
    property TagBackColor: TColor read FTagBackColor write SetTagBackColor;
    property TagPropForeColor: TColor read FTagPropForeColor write SetTagPropForeColor;
    property TagPropBackColor: TColor read FTagPropBackColor write SetTagPropBackColor;
    property ValueForeColor: TColor read FValueForeColor write SetValueForeColor;
    property ValueBackColor: TColor read FValueBackColor write SetValueBackColor;
    property SpecTagsForeColor: TColor read FSpecTagsForeColor write SSpecTagsForeColor;
    property SpecTagsBackColor: TColor read FSpecTagsBackColor write SSpecTagsBackColor;
    property TagBRForeColor: TColor read FBRTagForeColor write SBRTagForeColor;
    property TagBRBackColor: TColor read FBRTagBackColor write SBRTagBackColor;
    property UseBRTagHighlight: Boolean read FUseBRTagHighlight write SetUseBRTagHighlight;
    property UseSpecTagsHighlight: Boolean read FUseSpecTagsHighlight write SetUseSpecTagsHighlight;
  end;

function HTMLColorToColor(HTMLColor: string; var IsColor: Boolean): TColor;

implementation

function HTMLColorToColor(HTMLColor: string; var IsColor: Boolean): TColor;
var
  R, G, B, clStr: string;
begin
  IsColor := False;
  Result := clNone;
  R := '';
  G := '';
  B := '';
  if Length(HTMLColor) = 4 then
  begin
    if not (HTMLColor[1] = '#') then Exit;
    R := Copy(HTMLColor, 2, 1) + Copy(HTMLColor, 2, 1);
    G := Copy(HTMLColor, 3, 1) + Copy(HTMLColor, 3, 1);
    B := Copy(HTMLColor, 4, 1) + Copy(HTMLColor, 4, 1);
  end;
  if Length(HTMLColor) = 7 then
  begin
    if not (HTMLColor[1] = '#') then Exit;
    R := Copy(HTMLColor, 2, 2);
    G := Copy(HTMLColor, 4, 2);
    B := Copy(HTMLColor, 6, 2);
  end;
  clStr := '$00' + B + G + R;
  if Length(clStr) = 9 then
  try
    Result := StringToColor(clStr);
    IsColor := True;
  except
    IsColor := False;
    Result := clNone;
  end;
end;

(* TDropItem ********************************************************************************)

constructor TDropItem.Create;
begin
  inherited Create;
  FList := TStringList.Create;
end;

destructor TDropItem.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

procedure TDropItem.SetList(Value: TStringList);
begin
  FList.Assign(Value);
end;

(* TGsMemoHTMLCode ************************************************************************)

constructor TGsMemoHTMLCode.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCxMemo := nil;
  FCodeHighlight := True;
  FUseBRTagHighlight := True;
  FUseSpecTagsHighlight := True;
  FTagForeColor := clNavy;
  FTagBackColor := clWhite;
  FTagPropForeColor := clPurple;
  FTagPropBackColor := clWhite;
  FValueForeColor := clGray;
  FValueBackColor := clWhite;
  FSpecTagsForeColor := clGreen;
  FSpecTagsBackColor := clWhite;
  FBRTagForeColor := clRed;
  FBRTagBackColor := clWhite;
  FTagsList := TStringList.Create;
end;

destructor TGsMemoHTMLCode.Destroy;
begin
  ClearDropItems;
  FTagsList.Free;
  inherited Destroy;
end;

procedure TGsMemoHTMLCode.SetGsMemo(Value: TGsMemo);
begin
  FCxMemo := Value;
  if FCxMemo <> nil then
  begin
    FCxMemo.OnMemoEvents := MemoEvents;
    FCxMemo.OnDrawLine := MemoDrawLine;
    FCxMemo.OnDropListDown := DropListDown;
  end;
end;

procedure TGsMemoHTMLCode.SetCodeHighlight(Value: Boolean);
begin
  FCodeHighlight := Value;
  if FCxMemo <> nil then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetTagForeColor(Value: TColor);
begin
  FTagForeColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetTagBackColor(Value: TColor);
begin
  FTagBackColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetTagPropForeColor(Value: TColor);
begin
  FTagPropForeColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetTagPropBackColor(Value: TColor);
begin
  FTagPropBackColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetValueForeColor(Value: TColor);
begin
  FValueForeColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetValueBackColor(Value: TColor);
begin
  FValueBackColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SSpecTagsForeColor(Value: TColor);
begin
  FSpecTagsForeColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SSpecTagsBackColor(Value: TColor);
begin
  FSpecTagsBackColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SBRTagForeColor(Value: TColor);
begin
  FBRTagForeColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SBRTagBackColor(Value: TColor);
begin
  FBRTagBackColor := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetUseBRTagHighlight(Value: Boolean);
begin
  FUseBRTagHighlight := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.SetUseSpecTagsHighlight(Value: Boolean);
begin
  FUseSpecTagsHighlight := Value;
  if FCodeHighlight and (FCxMemo <> nil) then FCxMemo.Invalidate;
end;

procedure TGsMemoHTMLCode.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FCxMemo) then FCxMemo := nil;
end;

procedure TGsMemoHTMLCode.ShowColorHint(X, Y: Integer; HintColor: TColor);
var
  HintPos: TPoint;
  R: TRect;
begin
  if FCxMemo = nil then Exit;
  if FCxMemo.HintWindow <> nil then
  begin
    Exit;
    FCxMemo.HintWindow.Free;
    FCxMemo.HintWindow := nil;
  end;
  FCxMemo.HintWindow := THintWindow.Create(Self);
  HintPos := FCxMemo.ClientToScreen(Point(X, Y));
  FCxMemo.HintWindow.Color := HintColor;
  R := Rect(HintPos.x, HintPos.y, HintPos.x + 18, HintPos.y + 14);
  FCxMemo.HintWindow.ActivateHint(R, '');
end;

procedure TGsMemoHTMLCode.MemoEvents(Sender: TObject; Shift: TShiftState;
  KeyEvent: TGsKeyEvent; Key: Word; CaretPos, MousePos: TPoint);
var
  SLink: string;
  IsColor: Boolean;
  Cl: TColor;
  FChrW, FChrH, RNo, CNo, Y: Integer;
begin
  FChrH := FCxMemo.FontHeight;
  FChrW := FCxMemo.FontWidth;
  if (KeyEvent = k_Down) and (ssCtrl in Shift) then
  begin
    FCxMemo.GetRowColAtPos(MousePos.x + FCxMemo.ScrollPos_H * FChrW, MousePos.y +
      FCxMemo.ScrollPos_V * FChrH, RNo, CNo);
    SLink := UpperCase(FCxMemo.GetWordAtPos(CNo, RNo, Y));
    if (SLink <> '') and (Y <> -1) then
    begin
      IsColor := False;
      if SLink[1] = '#' then
        Cl := HTMLColorToColor(SLink, IsColor)
      else
        Cl := clNone;
      if Cl <> clNone then IsColor := True;
      if IsColor then
      begin
        ShowColorHint(MousePos.x + 4, MousePos.y + 4, Cl);
        FCxMemo.HideHintWindow := True;
      end else
        FCxMemo.HideHintWindow := False;
    end;
  end;
end;

procedure TGsMemoHTMLCode.ParseLine(LineText: string; Canvas: TCanvas);
var
  I, SaveI: Integer;
  S, SpecTags_S: String;
  R: TRect;
  hasTags: Boolean;
begin
  FCharCount := 0;
  FLastPos := 1;
  FStartPos := 0;
  FInTag := True;
  hasTags := False;
  SaveI := 1;
  for I := 1 to Length(LineText) do
  begin
    SaveI := I;
    case LineText[I] of
      '<': begin
         FInTag := True;
         FInValue := False;
         FFirstTag := True;
         FStartPos := I + 1;
         FCharCount := 0;
         FLastPos := I;
         hasTags := True;
         FTagNotClosed := True;
         Continue;
      end;
      '!': begin
        if FFirstTag and FInTag then FFirstTag := False;
        FCharCount := FCharCount + 1;
      end;
      ' ': if FFirstTag then
      begin
        FFirstTag := False;
        S := Copy(LineText, FStartPos, FCharCount);
        if FUseBRTagHighlight then
        begin
          if UpperCase(S) = 'BR' then
          begin
            Canvas.Font.Color := FBRTagForeColor;
            Canvas.Brush.Color := FBRTagBackColor;
          end else
          begin
            Canvas.Font.Color := FTagForeColor;
            Canvas.Brush.Color := FTagBackColor;
          end;
        end else
        begin
          Canvas.Font.Color := FTagForeColor;
          Canvas.Brush.Color := FTagBackColor;
        end;
        R := Rect(FCxMemo.FontWidth * (FLastPos), 0,
          FCxMemo.FontWidth * (FLastPos + FCharCount), FCxMemo.FontHeight);
        DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
        FCharCount := 0;
        FStartPos := I + 1;
        FLastPos := I;
        Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
        Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
        Continue;
      end else
        if FInTag then FCharCount := FCharCount + 1;
      '"': if FInTag and not FInValue then
      begin
        if FStartPos = 0 then
        begin
          FCharCount := I - 1;
          FLastPos := 0;
        end;
        FInValue := True;
        S := Copy(LineText, FStartPos, FCharCount);
        Canvas.Font.Color := FTagPropForeColor;
        Canvas.Brush.Color := FTagPropBackColor;
        R := Rect(FCxMemo.FontWidth * (FLastPos), 0,
          FCxMemo.FontWidth * (FLastPos + FCharCount), FCxMemo.FontHeight);
        DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
        FCharCount := 0;
        FStartPos := I + 1;
        FLastPos := I;
        Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
        Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
        Continue;
      end else
        if FInTag and FInValue then
        begin
          if FStartPos = 0 then
          begin
            FCharCount := I - 1;
            FLastPos := 0;
          end;
          FInValue := False;
          S := Copy(LineText, FStartPos, FCharCount);
          Canvas.Font.Color := FValueForeColor;
          Canvas.Brush.Color := FValueBackColor;
          R := Rect(FCxMemo.FontWidth * (FLastPos), 0,
            FCxMemo.FontWidth * (FLastPos + FCharCount), FCxMemo.FontHeight);
          DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
          FCharCount := 0;
          FStartPos := I + 1;
          FLastPos := I;
          Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
          Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
          Continue;
        end;
      '>': begin
         FInTag := False;
         FTagNotClosed := False;
         if FStartPos = 0 then
         begin
           FCharCount := I - 1;
           FLastPos := 0;
         end;
         S := Copy(LineText, FStartPos, FCharCount);
         if FFirstTag then
         begin
           FFirstTag := False;
           if FUseBRTagHighlight then
           begin
             if UpperCase(S) = 'BR' then
             begin
               Canvas.Font.Color := FBRTagForeColor;
               Canvas.Brush.Color := FBRTagBackColor;
             end else
             begin
               Canvas.Font.Color := FTagForeColor;
               Canvas.Brush.Color := FTagBackColor;
             end;
           end else
           begin
             Canvas.Font.Color := FTagForeColor;
             Canvas.Brush.Color := FTagBackColor;
           end;
         end else
         begin
           Canvas.Font.Color := FTagPropForeColor;
           Canvas.Brush.Color := FTagPropBackColor;
         end;
         R := Rect(FCxMemo.FontWidth * FLastPos, 0,
           FCxMemo.FontWidth * (FLastPos + FCharCount), FCxMemo.FontHeight);
         DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
         FCharCount := 0;
         FStartPos := I + 1;
         FLastPos := I;
         Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
         Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
         Continue;
      end else
        if FInTag then FCharCount := FCharCount + 1;
    end;
  end;
  if hasTags or FTagNotClosed then
  begin
    if FStartPos = 0 then
    begin
      FCharCount := SaveI - 1;
      FLastPos := 0;
    end;
    S := Copy(LineText, FStartPos, FCharCount);
    if FFirstTag then
    begin
      Canvas.Font.Color := FTagForeColor;
      Canvas.Brush.Color := FTagBackColor;
    end else
    begin
      Canvas.Font.Color := FTagPropForeColor;
      Canvas.Brush.Color := FTagPropBackColor;
    end;
    R := Rect(FCxMemo.FontWidth * FLastPos, 0,
      FCxMemo.FontWidth * (FLastPos + FCharCount), FCxMemo.FontHeight);
    DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
    Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
    Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
  end;
  if FUseSpecTagsHighlight then
  begin
    Canvas.Font.Color := FSpecTagsForeColor;
    Canvas.Brush.Color := FSpecTagsbackColor;
    SpecTags_S := UpperCase(LineText);
    I := Pos('&NBSP;', SpecTags_S);
    SaveI := 0;
    while I > 0 do
    begin
      SaveI := SaveI + I;
      S := Copy(LineText, SaveI, 6);
      Delete(SpecTags_S, 1, I + 5);
      R := Rect(FCxMemo.FontWidth * (SaveI - 1), 0,
        FCxMemo.FontWidth * (SaveI + 5), FCxMemo.FontHeight);
      DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
      I := Pos('&NBSP;', SpecTags_S);
      SaveI := SaveI + 5;
    end;
    SpecTags_S := UpperCase(LineText);
    I := Pos('&LT;', SpecTags_S);
    SaveI := 0;
    while I > 0 do
    begin
      SaveI := SaveI + I;
      S := Copy(LineText, SaveI, 4);
      Delete(SpecTags_S, 1, I + 3);
      R := Rect(FCxMemo.FontWidth * (SaveI - 1), 0,
        FCxMemo.FontWidth * (SaveI + 3), FCxMemo.FontHeight);
      DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
      I := Pos('&LT;', SpecTags_S);
      SaveI := SaveI + 3;
    end;
    SpecTags_S := UpperCase(LineText);
    I := Pos('&GT;', SpecTags_S);
    SaveI := 0;
    while I > 0 do
    begin
      SaveI := SaveI + I;
      S := Copy(LineText, SaveI, 4);
      Delete(SpecTags_S, 1, I + 3);
      R := Rect(FCxMemo.FontWidth * (SaveI - 1), 0,
        FCxMemo.FontWidth * (SaveI + 3), FCxMemo.FontHeight);
      DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
      I := Pos('&GT;', SpecTags_S);
      SaveI := SaveI + 3;
    end;
    SpecTags_S := UpperCase(LineText);
    I := Pos('&QUOT;', SpecTags_S);
    SaveI := 0;
    while I > 0 do
    begin
      SaveI := SaveI + I;
      S := Copy(LineText, SaveI, 6);
      Delete(SpecTags_S, 1, I + 5);
      R := Rect(FCxMemo.FontWidth * (SaveI - 1), 0,
        FCxMemo.FontWidth * (SaveI + 5), FCxMemo.FontHeight);
      DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
      I := Pos('&QUOT;', SpecTags_S);
      SaveI := SaveI + 5;
    end;
    SpecTags_S := UpperCase(LineText);
    I := Pos('&COPY;', SpecTags_S);
    SaveI := 0;
    while I > 0 do
    begin
      SaveI := SaveI + I;
      S := Copy(LineText, SaveI, 6);
      Delete(SpecTags_S, 1, I + 5);
      R := Rect(FCxMemo.FontWidth * (SaveI - 1), 0,
        FCxMemo.FontWidth * (SaveI + 5), FCxMemo.FontHeight);
      DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
      I := Pos('&COPY;', SpecTags_S);
      SaveI := SaveI + 5;
    end;
    SpecTags_S := UpperCase(LineText);
    I := Pos('&AMP;', SpecTags_S);
    SaveI := 0;
    while I > 0 do
    begin
      SaveI := SaveI + I;
      S := Copy(LineText, SaveI, 5);
      Delete(SpecTags_S, 1, I + 4);
      R := Rect(FCxMemo.FontWidth * (SaveI - 1), 0,
        FCxMemo.FontWidth * (SaveI + 4), FCxMemo.FontHeight);
      DrawText(Canvas.Handle, PChar(S), Length(S), R, DT_DRAWLINE);
      I := Pos('&AMP;', SpecTags_S);
      SaveI := SaveI + 4;
    end;
    Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
    Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
  end;
end;

procedure TGsMemoHTMLCode.MemoDrawLine(Sender: TObject; LineText: String; LineIndxe: Integer;
  LineGlyph: TBitmap; TextRect: TRect; var AllowSelfDraw: Boolean);
begin
  if FCxMemo = nil then Exit;
  AllowSelfDraw := not FCodeHighlight;
  if AllowSelfDraw then Exit;
  if LineIndxe = FCxMemo.TopLine then FTagNotClosed := False;
  LineGlyph.Canvas.Brush.Color := FCxMemo.EnvironmentOptions.TextBackground;
  LineGlyph.Canvas.Font.Color := FCxMemo.EnvironmentOptions.TextColor;
  DrawText(LineGlyph.Canvas.Handle, PChar(LineText), Length(LineText), TextRect, DT_DRAWLINE);
  ParseLine(LineText, LineGlyph.Canvas);
end;

function FindTagBegin(SLine: String; StartPos: Integer): Integer;
var
  I: Integer;
begin
  Result := -1;
  if (StartPos < 1) or (StartPos > Length(SLine)) then Exit;
  for I := StartPos downto 1 do
    if SLine[I] = '<' then
    begin
      Result := I;
      Break;
    end;
end;

function FindTagEnd(SLine: String; StartPos: Integer): Integer;
var
  I: Integer;
begin
  Result := -1;
  if (StartPos < 1) or (StartPos > Length(SLine)) then Exit;
  for I := StartPos to Length(SLine) do
    if SLine[I] = '>' then
    begin
      Result := I;
      Break;
    end;
end;

procedure TGsMemoHTMLCode.DropListDown(Sender: TObject);
var
  X, Y, tgB, tgE, I, I2, XBg: Integer;
  S, SLine, Tag: String;
  InTag: Boolean;
begin
  FCxMemo.DropListItems.Clear;
  X := FCxMemo.CaretPos_V;
  Y := FCxMemo.CaretPos_H;
  Tag := '';
  if X <= pred(FCxMemo.Lines.Count) then
  begin
    S := FCxMemo.Lines[X];
    tgB := FindTagBegin(S, Y + 1);
    tgE := FindTagEnd(S, Y + 1);
    if tgE > -1 then if (tgE < Y) and (tgE > tgB) then Exit;
    if (tgB > -1) and (tgB < Y + 1) then
    begin
      if S[tgB + 1] = '/' then Exit;
      Tag := FCxMemo.GetWordAtPos(tgB + 1, X, I);
      if (Y + 1) <= Length(S) then
      begin
        if (S[Y] in ['a'..'z', 'A'..'Z', ' ']) and (S[Y + 1] in [' ', '>']) then
          UpdateDropListForTag(Tag);
      end else
        if not (S[Y] in ['a'..'z', 'A'..'Z', ' ']) then
          UpdateDropListForTag(Tag);
      if S[Y] = '<' then UpdateDropList_AllTags;    
      Exit;
    end;
    InTag := False;
    SLine := FCxMemo.Lines[X];
    if tgB = -1 then
    begin
      if X = 0 then Exit;
      for I := X - 1 downto 0 do
      begin
        if InTag then Break;
        S := FCxMemo.Lines[I];
        for I2 := Length(S) downto 1 do
        begin
          if S[I2] = '>' then Exit;
          if S[I2] = '<' then
          begin
            InTag := True;
            if S[I2 + 1] = '/' then Exit;
            Tag := FCxMemo.GetWordAtPos(I2 + 1, I, XBg);
            if (Y + 1) <= Length(SLine) then
            begin
              if (SLine[Y] in ['a'..'z', 'A'..'Z', ' ']) and (SLine[Y + 1] in [' ', '>']) then
                UpdateDropListForTag(Tag);
            end else
              if (Y <= Length(SLine)) and (Y > 0) then
                if (SLine[Y] in ['a'..'z', 'A'..'Z', ' ']) then
                  UpdateDropListForTag(Tag);
            Break;
          end;
        end;
      end;
    end;
  end;
end;

procedure TGsMemoHTMLCode.UpdateDropListForTag(TagStr: String);
var
  I: Integer;
  DropItem: TDropItem;
begin
  I := FTagsList.IndexOf(UpperCase(TagStr));
  if I = -1 then Exit;
  if FTagsList.Objects[I] = nil then Exit;
  DropItem := TDropItem(FTagsList.Objects[I]);
  FCxMemo.DropListItems.Assign(DropItem.SubItems);
end;

procedure TGsMemoHTMLCode.UpdateDropList_AllTags;
var
  I: Integer;
  S, ItemS: String;
begin
  for I := 0 to pred(FTagsList.Count) do
  begin
    S := FTagsList[I];
    ItemS := '<' + LowerCase(S) + '>|HTML tag|' + LowerCase(S) + ' ';
    FCxMemo.DropListItems.Add(ItemS);
  end;
  FCxMemo.DropListItems.Add('<!--|HTML comment|!-- ');
end;

procedure TGsMemoHTMLCode.ClearDropItems;
var
  I: Integer;
begin
  I := pred(FTagsList.Count);
  if I = -1 then Exit;
  while I > -1 do
  begin
    if FTagsList.Objects[I] <> nil then FTagsList.Objects[I].Free;
    FTagsList.Delete(I);
    I := pred(FTagsList.Count);
  end;
end;

function TGsMemoHTMLCode.AddDropItem(DropItem: TDropItem): Integer;
begin
  if DropItem = nil then
    Result := -1
  else
    Result := FTagsList.AddObject(UpperCase(DropItem.Text), DropItem);
end;

procedure TGsMemoHTMLCode.DeleteDropItem(DropItemIndex: Integer);
begin
  if DropItemIndex > pred(FTagsList.Count) then Exit;
  FTagsList.Objects[DropItemIndex].Free;
  FTagsList.Delete(DropItemIndex);
end;

procedure TGsMemoHTMLCode.GetDropItemsFromDTDFile(const FileName: String);
var
  I: Integer;
  Tag, S: String;
  SList: TStringList;
  DI: TDropItem;
begin
  if not FileExists(FileName) then Exit;
  ClearDropItems;
  SList := TStringList.Create;
  SList.LoadFromFile(FileName);
  Tag := '';
  DI := nil;
  for I := 0 to pred(SList.Count) do
  begin
    S := SList[I];
    if Length(S) < 2 then Continue;
    if S[1] = '#' then
    begin
      if DI <> nil then AddDropItem(DI);
      Delete(S, 1, 1);
      Tag := S;
      DI := TDropItem.Create;
      DI.Text := S;
      Continue;
    end;
    if S[1] = Chr(9) then
    begin
      Delete(S, 1, 1);
      if DI = nil then Continue;
      if DI.Text = Tag then DI.SubItems.Add(S);
    end;
  end
end;

end.



